import jax
import jax.numpy as jnp
import jax.random as random
import warnings

from typing import Set


class Complement:
    '''Base class for approximate logical complement operations.'''
    
    def __call__(self, x):
        raise NotImplementedError


class StandardComplement(Complement):
    '''The standard approximate logical complement given by x -> 1 - x.'''
    
    def __call__(self, x):
        return 1.0 - x


class TNorm:
    '''Base class for fuzzy differentiable t-norms.'''
    
    def norm(self, x, y):
        '''Elementwise t-norm of x and y.'''
        raise NotImplementedError
    
    def norms(self, x, axis):
        '''T-norm computed for tensor x along axis.'''
        raise NotImplementedError
        

class ProductTNorm(TNorm):
    '''Product t-norm given by the expression (x, y) -> x * y.'''
    
    def norm(self, x, y):
        return x * y
    
    def norms(self, x, axis):
        return jnp.prod(x, axis=axis)


class GodelTNorm(TNorm):
    '''Godel t-norm given by the expression (x, y) -> min(x, y).'''
    
    def norm(self, x, y):
        return jnp.minimum(x, y)
    
    def norms(self, x, axis):
        return jnp.min(x, axis=axis)


class LukasiewiczTNorm(TNorm):
    '''Lukasiewicz t-norm given by the expression (x, y) -> max(x + y - 1, 0).'''
    
    def norm(self, x, y):
        return jax.nn.relu(x + y - 1.0)
    
    def norms(self, x, axis):
        return jax.nn.relu(jnp.sum(x - 1.0, axis=axis) + 1.0)
    
    
class FuzzyLogic:
    '''A class representing fuzzy logic in JAX.
    
    Functionality can be customized by either providing a tnorm as parameters, 
    or by overriding its methods.
    '''
    
    def __init__(self, tnorm: TNorm=ProductTNorm(),
                 complement: Complement=StandardComplement(),
                 weight: float=10.0,
                 debias: Set[str]={},
                 eps: float=1e-10,
                 verbose: bool=False):
        '''Creates a new fuzzy logic in Jax.
        
        :param tnorm: fuzzy operator for logical AND
        :param complement: fuzzy operator for logical NOT
        :param weight: a concentration parameter (larger means better accuracy)
        :param error: an error parameter (e.g. floor) (smaller means better accuracy)
        :param debias: which functions to de-bias approximate on forward pass
        :param eps: small positive float to mitigate underflow
        :param verbose: whether to dump replacements and other info to console
        '''
        self.tnorm = tnorm
        self.complement = complement
        self.weight = float(weight)
        self.debias = debias
        self.eps = eps
        self.verbose = verbose
    
    def summarize_hyperparameters(self):
        print(f'model relaxation:\n'
              f'    tnorm         ={type(self.tnorm).__name__}\n'
              f'    complement    ={type(self.complement).__name__}\n'
              f'    sigmoid_weight={self.weight}\n'
              f'    cpfs_to_debias={self.debias}\n'
              f'    underflow_tol ={self.eps}')
        
    # ===========================================================================
    # logical operators
    # ===========================================================================
     
    def And(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'a ^ b --> tnorm(a, b).', stacklevel=2)
        
        _and = self.tnorm.norm
        
        def _jax_wrapped_calc_and_approx(a, b, param):
            return _and(a, b)
        
        return _jax_wrapped_calc_and_approx, None
    
    def Not(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          '~a --> 1 - a', stacklevel=2)
        
        _not = self.complement
        
        def _jax_wrapped_calc_not_approx(x, param):
            return _not(x)
        
        return _jax_wrapped_calc_not_approx, None
    
    def Or(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'a or b --> tconorm(a, b).', stacklevel=2)
            
        _not = self.complement
        _and = self.tnorm.norm
        
        def _jax_wrapped_calc_or_approx(a, b, param):
            return _not(_and(_not(a), _not(b)))
        
        return _jax_wrapped_calc_or_approx, None

    def xor(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'a xor b --> (a or b) ^ (a ^ b).', stacklevel=2)
        
        _not = self.complement
        _and = self.tnorm.norm
        
        def _jax_wrapped_calc_xor_approx(a, b, param):
            _or = _not(_and(_not(a), _not(b)))
            return _and(_or(a, b), _not(_and(a, b)))
        
        return _jax_wrapped_calc_xor_approx, None
        
    def implies(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'a => b --> ~a ^ b', stacklevel=2)
        
        _not = self.complement
        _and = self.tnorm.norm
        
        def _jax_wrapped_calc_implies_approx(a, b, param):
            return _not(_and(a, _not(b)))
        
        return _jax_wrapped_calc_implies_approx, None
    
    def equiv(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'a <=> b --> (a => b) ^ (b => a)', stacklevel=2)
        
        _not = self.complement
        _and = self.tnorm.norm
        
        def _jax_wrapped_calc_equiv_approx(a, b, param):
            atob = _not(_and(a, _not(b)))
            btoa = _not(_and(b, _not(a)))
            return _and(atob, btoa)
        
        return _jax_wrapped_calc_equiv_approx, None
    
    def forall(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'forall(a) --> tnorm(a[1], tnorm(a[2], ...))', 
                          stacklevel=2)
        
        _forall = self.tnorm.norms
        
        def _jax_wrapped_calc_forall_approx(x, axis, param):
            return _forall(x, axis=axis)
        
        return _jax_wrapped_calc_forall_approx, None
    
    def exists(self):
        _not = self.complement
        jax_forall, jax_param = self.forall()
        
        def _jax_wrapped_calc_exists_approx(x, axis, param):
            return _not(jax_forall(_not(x), axis, param))
        
        return _jax_wrapped_calc_exists_approx, jax_param
    
    # ===========================================================================
    # comparison operators
    # ===========================================================================
     
    def greaterEqual(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'a >= b --> sigmoid(a - b)', stacklevel=2)
            
        debias = 'greaterEqual' in self.debias
        
        def _jax_wrapped_calc_geq_approx(a, b, param):
            sample = jax.nn.sigmoid(param * (a - b))
            if debias:
                hard_sample = jnp.greater_equal(a, b)
                sample += jax.lax.stop_gradient(hard_sample - sample)
            return sample
        
        tags = ('weight', 'greaterEqual')
        new_param = (tags, self.weight)
        return _jax_wrapped_calc_geq_approx, new_param
    
    def greater(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'a > b --> sigmoid(a - b)', stacklevel=2)
            
        debias = 'greater' in self.debias
        
        def _jax_wrapped_calc_gre_approx(a, b, param):
            sample = jax.nn.sigmoid(param * (a - b))
            if debias:
                hard_sample = jnp.greater(a, b)
                sample += jax.lax.stop_gradient(hard_sample - sample)
            return sample
        
        tags = ('weight', 'greater')
        new_param = (tags, self.weight)
        return _jax_wrapped_calc_gre_approx, new_param
    
    def lessEqual(self):
        jax_geq, jax_param = self.greaterEqual()
        
        def _jax_wrapped_calc_leq_approx(a, b, param):
            return jax_geq(-a, -b, param)
        
        return _jax_wrapped_calc_leq_approx, jax_param
    
    def less(self):
        jax_gre, jax_param = self.greater()

        def _jax_wrapped_calc_less_approx(a, b, param):
            return jax_gre(-a, -b, param)
        
        return _jax_wrapped_calc_less_approx, jax_param

    def equal(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'a == b --> sech^2(b - a)', stacklevel=2)
            
        debias = 'equal' in self.debias
        
        def _jax_wrapped_calc_equal_approx(a, b, param):
            sample = 1.0 - jnp.square(jnp.tanh(param * (b - a)))
            if debias:
                hard_sample = jnp.equal(a, b)
                sample += jax.lax.stop_gradient(hard_sample - sample)
            return sample
        
        tags = ('weight', 'equal')
        new_param = (tags, self.weight)
        return _jax_wrapped_calc_equal_approx, new_param
    
    def notEqual(self):
        _not = self.complement
        jax_eq, jax_param = self.equal()
        
        def _jax_wrapped_calc_neq_approx(a, b, param):
            return _not(jax_eq(a, b, param))

        return _jax_wrapped_calc_neq_approx, jax_param
        
    # ===========================================================================
    # special functions
    # ===========================================================================
     
    def signum(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'signum(x) --> tanh(x)', stacklevel=2)
            
        debias = 'signum' in self.debias
        
        def _jax_wrapped_calc_signum_approx(x, param):
            sample = jnp.tanh(param * x)
            if debias:
                hard_sample = jnp.sign(x)
                sample += jax.lax.stop_gradient(hard_sample - sample)
            return sample
        
        tags = ('weight', 'signum')
        new_param = (tags, self.weight)
        return _jax_wrapped_calc_signum_approx, new_param
    
    # def _sawtooth(self, x, d):
    #     pi = jnp.pi
    #     trg = 1.0 - 2.0 * jnp.arccos((1.0 - d) * jnp.sin(pi * (x - 0.5))) / pi
    #     sqr = 2.0 * jnp.arctan(jnp.sin(pi * x) / d) / pi
    #     swt = (1.0 + trg * sqr) / 2.0
    #     return swt        
    #
    # def floor(self):
    #     warnings.warn('Using the replacement rule: '
    #                   'floor(x) --> x - sawtooth(x), where sawtooth is a '
    #                   'trigonometric approximation of the sawtooth function',
    #                   stacklevel=2)
    #     debias = 'floor' in self.debias
    #
    #     def _jax_wrapped_calc_floor_approx(x, param):
    #         sample = x - self._sawtooth(x, param)
    #         if debias:
    #             hard_sample = jnp.floor(x)
    #             sample += jax.lax.stop_gradient(hard_sample - sample)
    #         return sample
    #
    #     tags = ('error', 'floor')
    #     new_param = (tags, self.error)
    #     return _jax_wrapped_calc_floor_approx, new_param
    #
    def ceil(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'ceil(x) --> ceil(x - 0.5) + step(x - 0.5), '
                          'where step is a smooth approximation of the step function')
            
        debias = 'ceil' in self.debias
        
        def _jax_wrapped_calc_ceil_approx(x, param):
            diff = (x - 0.5) - jnp.floor(x - 0.5)
            arg = jnp.log1p((2.0 * diff - 1.0) / (1.0 - diff) + self.eps)            
            sample = jnp.ceil(x - 0.5) + jax.nn.sigmoid(arg * param)
            if debias:
                hard_sample = jnp.ceil(x)
                sample += jax.lax.stop_gradient(hard_sample - sample)
            return sample
        
        tags = ('weight', 'ceil')
        new_param = (tags, self.weight)
        return _jax_wrapped_calc_ceil_approx, new_param
    
    def floor(self):
        jax_ceil, jax_param = self.ceil()
        
        def _jax_wrapped_calc_floor_approx(x, param):
            return -jax_ceil(-x, param)
        
        return _jax_wrapped_calc_floor_approx, jax_param
        
    def round(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'round(x) --> x', stacklevel=2)
        
        def _jax_wrapped_calc_round_approx(x, param):
            return x
        
        return _jax_wrapped_calc_round_approx, None
    
    def mod(self):
        jax_floor, jax_param = self.floor()
        
        def _jax_wrapped_calc_mod_approx(x, y, param):
            return x - y * jax_floor(x / y, param)
        
        return _jax_wrapped_calc_mod_approx, jax_param
    
    def floorDiv(self):
        jax_floor, jax_param = self.floor()
        
        def _jax_wrapped_calc_mod_approx(x, y, param):
            return jax_floor(x / y, param)
        
        return _jax_wrapped_calc_mod_approx, jax_param
    
    def sqrt(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'sqrt(x) --> sqrt(x + eps)', stacklevel=2)
        
        def _jax_wrapped_calc_sqrt_approx(x, param):
            return jnp.sqrt(x + self.eps)
        
        return _jax_wrapped_calc_sqrt_approx, None
    
    # ===========================================================================
    # indexing
    # ===========================================================================
     
    @staticmethod
    def _literals(shape, axis):
        literals = jnp.arange(shape[axis])
        literals = literals[(...,) + (jnp.newaxis,) * (len(shape) - 1)]
        literals = jnp.moveaxis(literals, source=0, destination=axis)
        literals = jnp.broadcast_to(literals, shape=shape)
        return literals
    
    def argmax(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          f'argmax(x) --> sum(i * softmax(x[i]))', stacklevel=2)
            
        debias = 'argmax' in self.debias
        
        def _jax_wrapped_calc_argmax_approx(x, axis, param):
            prob_max = jax.nn.softmax(param * x, axis=axis)
            literals = FuzzyLogic._literals(prob_max.shape, axis=axis)
            sample = jnp.sum(literals * prob_max, axis=axis)
            if debias:
                hard_sample = jnp.argmax(x, axis=axis)
                sample += jax.lax.stop_gradient(hard_sample - sample)
            return sample
        
        tags = ('weight', 'argmax')
        new_param = (tags, self.weight)
        return _jax_wrapped_calc_argmax_approx, new_param
    
    def argmin(self):
        jax_argmax, jax_param = self.argmax()
        
        def _jax_wrapped_calc_argmin_approx(x, axis, param):
            return jax_argmax(-x, axis, param)
        
        return _jax_wrapped_calc_argmin_approx, jax_param
    
    # ===========================================================================
    # control flow
    # ===========================================================================
     
    def If(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'if c then a else b --> c * a + (1 - c) * b', stacklevel=2)
        
        def _jax_wrapped_calc_if_approx(c, a, b, param):
            return c * a + (1.0 - c) * b
        
        return _jax_wrapped_calc_if_approx, None
    
    def Switch(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'switch(pred) { cases } --> sum(cases[i] * (pred == i))',
                          stacklevel=2)   
            
        debias = 'Switch' in self.debias
        
        def _jax_wrapped_calc_switch_approx(pred, cases, param):
            pred = jnp.broadcast_to(pred[jnp.newaxis, ...], shape=cases.shape)
            literals = FuzzyLogic._literals(cases.shape, axis=0)
            proximity = -jnp.abs(pred - literals)
            soft_case = jax.nn.softmax(param * proximity, axis=0)
            sample = jnp.sum(cases * soft_case, axis=0)
            if debias:
                hard_case = jnp.argmax(proximity, axis=0)[jnp.newaxis, ...]      
                hard_sample = jnp.take_along_axis(cases, hard_case, axis=0)[0, ...]
                sample += jax.lax.stop_gradient(hard_sample - sample)
            return sample
        
        tags = ('weight', 'switch')
        new_param = (tags, self.weight)
        return _jax_wrapped_calc_switch_approx, new_param
    
    # ===========================================================================
    # random variables
    # ===========================================================================
     
    def _gumbel_softmax(self, key, prob):
        Gumbel01 = random.gumbel(key=key, shape=prob.shape)
        sample = Gumbel01 + jnp.log1p(prob + self.eps - 1.0)
        return sample
        
    def bernoulli(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'Bernoulli(p) --> Gumbel-softmax(p)', stacklevel=2)
        
        jax_gs = self._gumbel_softmax
        jax_argmax, jax_param = self.argmax()
        
        def _jax_wrapped_calc_switch_approx(key, prob, param):
            prob = jnp.stack([1.0 - prob, prob], axis=-1)
            sample = jax_gs(key, prob)
            sample = jax_argmax(sample, -1, param)
            return sample
        
        return _jax_wrapped_calc_switch_approx, jax_param
    
    def discrete(self):
        if self.verbose:
            warnings.warn('Using the replacement rule: '
                          'Discrete(p) --> Gumbel-softmax(p)', stacklevel=2)
        
        jax_gs = self._gumbel_softmax
        jax_argmax, jax_param = self.argmax()
        
        def _jax_wrapped_calc_discrete_approx(key, prob, param):
            sample = jax_gs(key, prob) 
            sample = jax_argmax(sample, -1, param)
            return sample
        
        return _jax_wrapped_calc_discrete_approx, jax_param
        

# UNIT TESTS
logic = FuzzyLogic()
w = 100.0


def _test_logical():
    print('testing logical')
    _and, _ = logic.And()
    _not, _ = logic.Not()
    _gre, _ = logic.greater()
    _or, _ = logic.Or()
    _if, _ = logic.If()
    
    # https://towardsdatascience.com/emulating-logical-gates-with-a-neural-network-75c229ec4cc9
    def test_logic(x1, x2):
        q1 = _and(_gre(x1, 0, w), _gre(x2, 0, w), w)
        q2 = _and(_not(_gre(x1, 0, w), w), _not(_gre(x2, 0, w), w), w)
        cond = _or(q1, q2, w)
        pred = _if(cond, +1, -1, w)
        return pred
    
    x1 = jnp.asarray([1, 1, -1, -1, 0.1, 15, -0.5]).astype(float)
    x2 = jnp.asarray([1, -1, 1, -1, 10, -30, 6]).astype(float)
    print(test_logic(x1, x2))


def _test_indexing():
    print('testing indexing')
    _argmax, _ = logic.argmax()
    _argmin, _ = logic.argmax()

    def argmaxmin(x):
        amax = _argmax(x, 0, w)
        amin = _argmin(x, 0, w)
        return amax, amin
        
    values = jnp.asarray([2., 3., 5., 4.9, 4., 1., -1., -2.])
    amax, amin = argmaxmin(values)
    print(amax)
    print(amin)


def _test_control():
    print('testing control')
    _switch, _ = logic.Switch()
    
    pred = jnp.asarray(jnp.linspace(0, 2, 10))
    case1 = jnp.asarray([-10.] * 10)
    case2 = jnp.asarray([1.5] * 10)
    case3 = jnp.asarray([10.] * 10)
    cases = jnp.asarray([case1, case2, case3])
    print(_switch(pred, cases, w))


def _test_random():
    print('testing random')
    key = random.PRNGKey(42)
    _bernoulli, _ = logic.bernoulli()
    _discrete, _ = logic.discrete()
    
    def bern(n):
        prob = jnp.asarray([0.3] * n)
        sample = _bernoulli(key, prob, w)
        return sample
    
    samples = bern(5000)
    print(jnp.mean(samples))
    
    def disc(n):
        prob = jnp.asarray([0.1, 0.4, 0.5])
        prob = jnp.tile(prob, (n, 1))
        sample = _discrete(key, prob, w)
        return sample
        
    samples = disc(5000)
    samples = jnp.round(samples)
    print([jnp.mean(samples == i) for i in range(3)])


def _test_rounding():
    print('testing rounding')
    _floor, _ = logic.floor()
    _ceil, _ = logic.ceil()
    _mod, _ = logic.mod()
    
    x = jnp.asarray([2.1, 0.5001, 1.99, -2.01, -3.2, -0.1, -1.01, 23.01, -101.99, 200.01])
    print(_floor(x, w))
    print(_ceil(x, w))
    print(_mod(x, 2.0, w))


if __name__ == '__main__':
    _test_logical()
    _test_indexing()
    _test_control()
    _test_random()
    _test_rounding()
    



from abc import ABCMeta, abstractmethod
import os
from typing import Dict


class InstanceGenerator(metaclass=ABCMeta):
    
    @abstractmethod
    def get_env_path(self) -> str:
        pass

    @abstractmethod
    def get_domain_name(self) -> str:
        pass

    @abstractmethod
    def sample_instance(self, params: Dict[str, object]) -> Dict[str, object]:
        pass
    
    def generate_instance(self, inst_name: str, params: Dict[str, object]) -> str:
        instance = self.sample_instance(params)
        objects = instance['objects']
        nonfluents = instance['non-fluents']
        states = instance['init-states']
        horizon = instance['horizon']
        discount = instance['discount']
        actions = instance['max-nondef-actions']
        
        domain_name = self.get_domain_name()
        
        # object lists
        objects_lines = [f'{key} : {{{", ".join(val)}}};' 
                         for (key, val) in objects.items()]
        
        # non-fluents list
        nonfluent_lines = []
        for (key, val) in nonfluents.items():
            if isinstance(val, bool):
                if val == True:
                    nonfluent_lines.append(f'{key};')
                else:
                    nonfluent_lines.append(f'~{key};')
            else:
                valstr = f'{val}'
                if 'e' in valstr or 'E' in valstr:
                    valstr = '{:.15f}'.format(val)
                nonfluent_lines.append(f'{key} = {valstr};')
        
        # init-state list
        states_lines = []
        for (key, val) in states.items():
            if isinstance(val, bool):
                if val == True:
                    states_lines.append(f'{key};')
                else:
                    states_lines.append(f'~{key};')
            else: 
                valstr = f'{val}'
                if 'e' in valstr or 'E' in valstr:
                    valstr = '{:.15f}'.format(val)
                states_lines.append(f'{key} = {valstr};')
        
        # generate non-fluents block
        value = f'non-fluents nf_{domain_name}_{inst_name}' + ' {'
        value += '\n\t' + f'domain = {domain_name};'
        if objects_lines:
            value += '\n\t' + 'objects {' + '\n\t\t'
            value += '\n\t\t'.join(objects_lines)
            value += '\n\t' + '};'
        if nonfluent_lines:
            value += '\n\t' + 'non-fluents {' + '\n\t\t'
            value += '\n\t\t'.join(nonfluent_lines)
            value += '\n\t' + '};'
        value += '\n' + '}'
        
        # generate instance block
        value += '\n' + f'instance inst_{domain_name}_{inst_name}' + ' {'
        value += '\n\t' + f'domain = {domain_name};'
        value += '\n\t' + f'non-fluents = nf_{domain_name}_{inst_name};'
        if states_lines:
            value += '\n\t' + 'init-state {' + '\n\t\t'
            value += '\n\t\t'.join(states_lines)
            value += '\n\t' + '};'
        value += '\n\t' + f'max-nondef-actions = {actions};'
        value += '\n\t' + f'horizon = {horizon};'
        value += '\n\t' + f'discount = {discount};'
        value += '\n' + '}'
        return value
        
    def save_instance(self, instance: int, params: Dict[str, object], path=None) -> None:
        if path == None:
            path = os.path.dirname(os.path.abspath(__file__))
            path = os.path.join(path, self.get_env_path())

        dir = path
        path = os.path.join(path, f'instance{instance}.rddl')
        rddl = self.generate_instance(instance, params)
        isExist = os.path.exists(dir)
        if not isExist:
            # Create a new directory because it does not exist
            os.makedirs(dir)
        with open(path, 'w') as text_file:
            text_file.write(rddl)
        print(f'saved RDDL file to {path}.')
        


from bayes_opt import BayesianOptimization
from bayes_opt.util import UtilityFunction
from copy import deepcopy
import csv
import datetime
import jax
from multiprocessing import get_context
import numpy as np
import os
import time
from typing import Callable, Dict, Tuple
import warnings
warnings.filterwarnings('ignore')

from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxOfflineController
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxOnlineController
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxRDDLBackpropPlanner
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxStraightLinePlan
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxDeepReactivePolicy

# do this after imports to prevent it from being overwritten
np.seterr(all='warn')


# ===============================================================================
# 
# GENERIC TUNING MODULE
# 
# Currently contains three implementations:
# 1. straight line plan
# 2. re-planning
# 3. deep reactive policies
# 
# ===============================================================================
class JaxParameterTuning:
    '''A general-purpose class for tuning a Jax planner.'''
    
    def __init__(self, env: RDDLEnv,
                 hyperparams_dict: Dict[str, Tuple[float, float, Callable]],
                 train_epochs: int,
                 timeout_training: float,
                 timeout_tuning: float=np.inf,
                 eval_trials: int=5,
                 verbose: bool=True,
                 planner_kwargs: Dict={},
                 plan_kwargs: Dict={},
                 pool_context: str='spawn',
                 num_workers: int=1,
                 poll_frequency: float=0.2,
                 gp_iters: int=25,
                 acquisition=None,
                 gp_init_kwargs: Dict={},
                 gp_params: Dict={'n_restarts_optimizer': 10}) -> None:
        '''Creates a new instance for tuning hyper-parameters for Jax planners
        on the given RDDL domain and instance.
        
        :param env: the RDDLEnv describing the MDP to optimize
        :param hyperparams_dict: dictionary mapping name of each hyperparameter
        to a triple, where the first two elements are lower/upper bounds on the
        parameter value, and the last is a callable mapping the parameter to its
        RDDL equivalent
        :param train_epochs: the maximum number of iterations of SGD per 
        step or trial
        :param timeout_training: the maximum amount of time to spend training per
        trial/decision step (in seconds)
        :param timeout_tuning: the maximum amount of time to spend tuning 
        hyperparameters in general (in seconds)
        :param eval_trials: how many trials to perform independent training
        in order to estimate the return for each set of hyper-parameters
        :param verbose: whether to print intermediate results of tuning
        :param planner_kwargs: additional arguments to feed to the planner
        :param plan_kwargs: additional arguments to feed to the plan/policy
        :param pool_context: context for multiprocessing pool (defaults to 
        "spawn")
        :param num_workers: how many points to evaluate in parallel
        :param poll_frequency: how often (in seconds) to poll for completed
        jobs, necessary if num_workers > 1
        :param gp_iters: number of iterations of optimization
        :param acquisition: acquisition function for Bayesian optimizer
        :parm gp_init_kwargs: additional parameters to feed to Bayesian 
        during initialization  
        :param gp_params: additional parameters to feed to Bayesian optimizer 
        after initialization optimization
        '''
        
        self.env = env
        self.hyperparams_dict = hyperparams_dict
        self.train_epochs = train_epochs
        self.timeout_training = timeout_training
        self.timeout_tuning = timeout_tuning
        self.eval_trials = eval_trials
        self.verbose = verbose
        self.planner_kwargs = planner_kwargs
        self.plan_kwargs = plan_kwargs
        self.pool_context = pool_context
        self.num_workers = num_workers
        self.poll_frequency = poll_frequency
        self.gp_iters = gp_iters
        self.gp_init_kwargs = gp_init_kwargs
        self.gp_params = gp_params
        
        # create acquisition function
        self.acq_args = None
        if acquisition is None:
            num_samples = self.gp_iters * self.num_workers
            acquisition, self.acq_args = JaxParameterTuning._annealing_utility(num_samples)
        self.acquisition = acquisition
    
    def summarize_hyperparameters(self):
        print(f'hyperparameter optimizer parameters:\n'
              f'    tuned_hyper_parameters    ={self.hyperparams_dict}\n'
              f'    initialization_args       ={self.gp_init_kwargs}\n'
              f'    additional_args           ={self.gp_params}\n'
              f'    tuning_iterations         ={self.gp_iters}\n'
              f'    tuning_timeout            ={self.timeout_tuning}\n'
              f'    tuning_batch_size         ={self.num_workers}\n'
              f'    mp_pool_context_type      ={self.pool_context}\n'
              f'    mp_pool_poll_frequency    ={self.poll_frequency}\n'
              f'meta-objective parameters:\n'
              f'    planning_trials_per_iter  ={self.eval_trials}\n'
              f'    planning_iters_per_trial  ={self.train_epochs}\n'
              f'    planning_timeout_per_trial={self.timeout_training}\n'
              f'    acquisition_fn            ={type(self.acquisition).__name__}')
        if self.acq_args is not None:
            print(f'using default acquisition function:\n'
                  f'    utility_kind ={self.acq_args[0]}\n'
                  f'    initial_kappa={self.acq_args[1]}\n'
                  f'    kappa_decay  ={self.acq_args[2]}')
        
    @staticmethod
    def _annealing_utility(n_samples, n_delay_samples=0, kappa1=10.0, kappa2=1.0):
        kappa_decay = (kappa2 / kappa1) ** (1.0 / (n_samples - n_delay_samples))
        utility_fn = UtilityFunction(
            kind='ucb',
            kappa=kappa1,
            kappa_decay=kappa_decay,
            kappa_decay_delay=n_delay_samples)
        utility_args = ['ucb', kappa1, kappa_decay]
        return utility_fn, utility_args
    
    def _pickleable_objective_with_kwargs(self):
        raise NotImplementedError
    
    @staticmethod
    def _wrapped_evaluate(index, params, key, func, kwargs):
        target = func(params=params, kwargs=kwargs, key=key, index=index)
        pid = os.getpid()
        return index, pid, params, target

    def tune(self, key: jax.random.PRNGKey, filename: str,
             save_plot: bool=False) -> Dict[str, object]:
        '''Tunes the hyper-parameters for Jax planner, returns the best found.'''
        self.summarize_hyperparameters()
        
        start_time = time.time()
        
        # objective function
        objective = self._pickleable_objective_with_kwargs()
        evaluate = JaxParameterTuning._wrapped_evaluate
            
        # create optimizer
        hyperparams_bounds = {
            name: hparam[:2] 
            for (name, hparam) in self.hyperparams_dict.items()
        }
        optimizer = BayesianOptimization(
            f=None,  # probe() is not called
            pbounds=hyperparams_bounds,
            allow_duplicate_points=True,  # to avoid crash
            random_state=np.random.RandomState(key),
            **self.gp_init_kwargs
        )
        optimizer.set_gp_params(**self.gp_params)
        utility = self.acquisition
        
        # suggest initial parameters to evaluate
        num_workers = self.num_workers
        suggested, kappas = [], []
        for _ in range(num_workers):
            utility.update_params()
            probe = optimizer.suggest(utility)
            suggested.append(probe)  
            kappas.append(utility.kappa)
        
        # clear and prepare output file
        filename = self._filename(filename, 'csv')
        with open(filename, 'w', newline='') as file:
            writer = csv.writer(file)
            writer.writerow(
                ['pid', 'worker', 'iteration', 'target', 'best_target', 'kappa'] + \
                list(hyperparams_bounds.keys())
            )
                
        # start multiprocess evaluation
        worker_ids = list(range(num_workers))
        best_params, best_target = None, -np.inf
        
        for it in range(self.gp_iters): 
            
            # check if there is enough time left for another iteration
            elapsed = time.time() - start_time
            if elapsed >= self.timeout_tuning:
                print(f'global time limit reached at iteration {it}, aborting')
                break
            
            # continue with next iteration
            print('\n' + '*' * 25 + 
                  '\n' + f'[{datetime.timedelta(seconds=elapsed)}] ' + 
                  f'starting iteration {it}' + 
                  '\n' + '*' * 25)
            key, *subkeys = jax.random.split(key, num=num_workers + 1)
            rows = [None] * num_workers
            
            # create worker pool: note each iteration must wait for all workers
            # to finish before moving to the next
            with get_context(self.pool_context).Pool(processes=num_workers) as pool:
                
                # assign jobs to worker pool
                # - each trains on suggested parameters from the last iteration
                # - this way, since each job finishes asynchronously, these
                # parameters usually differ across jobs
                results = [
                    pool.apply_async(evaluate, worker_args + objective)
                    for worker_args in zip(worker_ids, suggested, subkeys)
                ]
            
                # wait for all workers to complete
                while results:
                    time.sleep(self.poll_frequency)
                    
                    # determine which jobs have completed
                    jobs_done = []
                    for (i, candidate) in enumerate(results):
                        if candidate.ready():
                            jobs_done.append(i)
                    
                    # get result from completed jobs
                    for i in jobs_done[::-1]:
                        
                        # extract and register the new evaluation
                        index, pid, params, target = results.pop(i).get()
                        optimizer.register(params, target)
                        
                        # update acquisition function and suggest a new point
                        utility.update_params()  
                        suggested[index] = optimizer.suggest(utility)
                        old_kappa = kappas[index]
                        kappas[index] = utility.kappa
                        
                        # transform suggestion back to natural space
                        rddl_params = {
                            name: pf(params[name])
                            for (name, (*_, pf)) in self.hyperparams_dict.items()
                        }
                        
                        # update the best suggestion so far
                        if target > best_target:
                            best_params, best_target = rddl_params, target
                        
                        # write progress to file in real time
                        rows[index] = [pid, index, it, target, best_target, old_kappa] + \
                                      list(rddl_params.values())
                        
            # write results of all processes in current iteration to file
            with open(filename, 'a', newline='') as file:
                writer = csv.writer(file)
                writer.writerows(rows)
        
        # print summary of results
        elapsed = time.time() - start_time
        print(f'summary of hyper-parameter optimization:\n'
              f'    time_elapsed         ={datetime.timedelta(seconds=elapsed)}\n'
              f'    iterations           ={it + 1}\n'
              f'    best_hyper_parameters={best_params}\n'
              f'    best_meta_objective  ={best_target}\n')
        
        if save_plot:
            self._save_plot(filename)
        return best_params

    def _filename(self, name, ext):
        domainName = self.env.model.domainName()
        instName = self.env.model.instanceName()
        domainName = ''.join(c for c in domainName if c.isalnum() or c == '_')
        instName = ''.join(c for c in instName if c.isalnum() or c == '_')
        filename = f'{name}_{domainName}_{instName}.{ext}'
        return filename
    
    def _save_plot(self, filename):
        try:
            import matplotlib.pyplot as plt
            from sklearn.manifold import MDS
        except Exception as e:
            warnings.warn(f'failed to import packages matplotlib or sklearn, '
                          f'aborting plot of search space\n'
                          f'{e}', stacklevel=2)
        else:
            data = np.loadtxt(filename, delimiter=',', dtype=object)
            data, target = data[1:, 3:], data[1:, 2]
            data = data.astype(np.float64)
            target = target.astype(np.float64)
            target = (target - np.min(target)) / (np.max(target) - np.min(target))
            embedding = MDS(n_components=2, normalized_stress='auto')
            data1 = embedding.fit_transform(data)
            sc = plt.scatter(data1[:, 0], data1[:, 1], c=target, s=4.,
                             cmap='seismic', edgecolor='gray',
                             linewidth=0.01, alpha=0.4)
            plt.colorbar(sc)
            plt.savefig(self._filename('gp_points', 'pdf'))
            plt.clf()
            plt.close()


# ===============================================================================
# 
# STRAIGHT LINE PLANNING
#
# ===============================================================================
def objective_slp(params, kwargs, key, index):
                    
    # transform hyper-parameters to natural space
    param_values = [
        pmap(params[name])
        for (name, (*_, pmap)) in kwargs['hyperparams_dict'].items()
    ]
    
    # unpack hyper-parameters
    if kwargs['wrapped_bool_actions']:
        std, lr, w, wa = param_values
    else:
        std, lr, w = param_values
        wa = None                      
    if kwargs['verbose']:
        print(f'[{index}] key={key}, std={std}, lr={lr}, w={w}, wa={wa}...', flush=True)
        
    # initialize planning algorithm
    planner = JaxRDDLBackpropPlanner(
        rddl=deepcopy(kwargs['rddl']),
        plan=JaxStraightLinePlan(
            initializer=jax.nn.initializers.normal(std),
            **kwargs['plan_kwargs']),
        optimizer_kwargs={'learning_rate': lr},
        **kwargs['planner_kwargs'])
    policy_hparams = {name: wa for name in kwargs['wrapped_bool_actions']}  
    model_params = {name: w for name in planner.compiled.model_params}
    
    # initialize policy
    key, subkey = jax.random.split(key)
    policy = JaxOfflineController(
        planner=planner,
        key=subkey,
        eval_hyperparams=policy_hparams,
        train_on_reset=True,
        epochs=kwargs['train_epochs'],
        train_seconds=kwargs['timeout_training'],
        model_params=model_params,
        policy_hyperparams=policy_hparams,
        verbose=0,
        tqdm_position=index)
    
    # initialize env for evaluation (need fresh copy to avoid concurrency)
    env = RDDLEnv(domain=kwargs['domain'],
                  instance=kwargs['instance'],
                  enforce_action_constraints=True)

    # perform training
    average_reward = 0.0
    for trial in range(kwargs['eval_trials']):
        key, subkey = jax.random.split(key)
        total_reward = policy.evaluate(env, seed=np.array(subkey)[0])['mean']
        if kwargs['verbose']:
            print(f'    [{index}] trial {trial + 1} key={subkey}, '
                  f'reward={total_reward}', flush=True)
        average_reward += total_reward / kwargs['eval_trials']        
    if kwargs['verbose']:
        print(f'[{index}] average reward={average_reward}', flush=True)
    return average_reward

        
def power_ten(x):
    return 10.0 ** x

    
class JaxParameterTuningSLP(JaxParameterTuning):
    
    def __init__(self, *args,
                 hyperparams_dict: Dict[str, Tuple[float, float, Callable]]={
                    'std': (-5., 2., power_ten),
                    'lr': (-5., 2., power_ten),
                    'w': (0., 5., power_ten),
                    'wa': (0., 5., power_ten)
                 },
                 **kwargs) -> None:
        '''Creates a new tuning class for straight line planners.
        
        :param *args: arguments to pass to parent class
        :param hyperparams_dict: same as parent class, but here must contain
        weight initialization (std), learning rate (lr), model weight (w), and
        action weight (wa) if wrap_sigmoid and boolean action fluents exist
        :param **kwargs: keyword arguments to pass to parent class
        '''
        
        super(JaxParameterTuningSLP, self).__init__(
            *args, hyperparams_dict=hyperparams_dict, **kwargs)
        
        # action parameters required if wrap_sigmoid and boolean action exists
        self.wrapped_bool_actions = []
        if self.plan_kwargs.get('wrap_sigmoid', True):
            for var in self.env.model.actions:
                if self.env.model.variable_ranges[var] == 'bool':
                    self.wrapped_bool_actions.append(var)
        if not self.wrapped_bool_actions:
            self.hyperparams_dict.pop('wa', None)
        
    def _pickleable_objective_with_kwargs(self):
        objective_fn = objective_slp
        
        # duplicate planner and plan keyword arguments must be removed
        plan_kwargs = self.plan_kwargs.copy()
        plan_kwargs.pop('initializer', None) 
               
        planner_kwargs = self.planner_kwargs.copy()
        planner_kwargs.pop('rddl', None)
        planner_kwargs.pop('plan', None)
        planner_kwargs.pop('optimizer_kwargs', None)
                    
        kwargs = {
            'rddl': self.env.model,
            'domain': self.env.domain_text,
            'instance': self.env.instance_text,
            'hyperparams_dict': self.hyperparams_dict,
            'timeout_training': self.timeout_training,
            'train_epochs': self.train_epochs,
            'planner_kwargs': planner_kwargs,
            'plan_kwargs': plan_kwargs,
            'verbose': self.verbose,
            'wrapped_bool_actions': self.wrapped_bool_actions,
            'eval_trials': self.eval_trials
        }
        return objective_fn, kwargs


# ===============================================================================
# 
# REPLANNING
#
# ===============================================================================
def objective_replan(params, kwargs, key, index):

    # transform hyper-parameters to natural space
    param_values = [
        pmap(params[name])
        for (name, (*_, pmap)) in kwargs['hyperparams_dict'].items()
    ]
    
    # unpack hyper-parameters
    if kwargs['wrapped_bool_actions']:
        std, lr, w, wa, T = param_values
    else:
        std, lr, w, T = param_values
        wa = None        
    if kwargs['verbose']:
        print(f'[{index}] key={key}, std={std}, lr={lr}, w={w}, wa={wa}, T={T}...', flush=True)

    # initialize planning algorithm
    planner = JaxRDDLBackpropPlanner(
        rddl=deepcopy(kwargs['rddl']),
        plan=JaxStraightLinePlan(
            initializer=jax.nn.initializers.normal(std),
            **kwargs['plan_kwargs']),
        rollout_horizon=T,
        optimizer_kwargs={'learning_rate': lr},
        **kwargs['planner_kwargs'])
    policy_hparams = {name: wa for name in kwargs['wrapped_bool_actions']}
    model_params = {name: w for name in planner.compiled.model_params}
    
    # initialize controller
    key, subkey = jax.random.split(key)
    policy = JaxOnlineController(
        planner=planner,
        key=subkey,
        eval_hyperparams=policy_hparams,
        warm_start=kwargs['use_guess_last_epoch'],
        epochs=kwargs['train_epochs'],
        train_seconds=kwargs['timeout_training'],
        model_params=model_params,
        policy_hyperparams=policy_hparams,
        verbose=0,
        tqdm_position=index)
    
    # initialize env for evaluation (need fresh copy to avoid concurrency)
    env = RDDLEnv(domain=kwargs['domain'],
                  instance=kwargs['instance'],
                  enforce_action_constraints=True)

    # perform training
    average_reward = 0.0
    for trial in range(kwargs['eval_trials']):
        key, subkey = jax.random.split(key)
        total_reward = policy.evaluate(env, seed=np.array(subkey)[0])['mean']
        if kwargs['verbose']:
            print(f'    [{index}] trial {trial + 1} key={subkey}, '
                  f'reward={total_reward}', flush=True)
        average_reward += total_reward / kwargs['eval_trials']        
    if kwargs['verbose']:
        print(f'[{index}] average reward={average_reward}', flush=True)
    return average_reward

    
class JaxParameterTuningSLPReplan(JaxParameterTuningSLP):
    
    def __init__(self,
                 *args,
                 hyperparams_dict: Dict[str, Tuple[float, float, Callable]]={
                    'std': (-5., 2., power_ten),
                    'lr': (-5., 2., power_ten),
                    'w': (0., 5., power_ten),
                    'wa': (0., 5., power_ten),
                    'T': (1, None, int)
                 },
                 use_guess_last_epoch: bool=True,
                 **kwargs) -> None:
        '''Creates a new tuning class for straight line planners.
        
        :param *args: arguments to pass to parent class
        :param hyperparams_dict: same as parent class, but here must contain
        weight initialization (std), learning rate (lr), model weight (w), 
        action weight (wa) if wrap_sigmoid and boolean action fluents exist, and
        lookahead horizon (T)
        :param use_guess_last_epoch: use the trained parameters from previous 
        decision to warm-start next decision
        :param **kwargs: keyword arguments to pass to parent class
        '''
        
        super(JaxParameterTuningSLPReplan, self).__init__(
            *args, hyperparams_dict=hyperparams_dict, **kwargs)
        
        self.use_guess_last_epoch = use_guess_last_epoch
        
        # set upper range of lookahead horizon to environment horizon
        if self.hyperparams_dict['T'][1] is None:
            self.hyperparams_dict['T'] = (1, self.env.horizon, int)
            
    def _pickleable_objective_with_kwargs(self):
        objective_fn = objective_replan
            
        # duplicate planner and plan keyword arguments must be removed
        plan_kwargs = self.plan_kwargs.copy()
        plan_kwargs.pop('initializer', None)
        
        planner_kwargs = self.planner_kwargs.copy()
        planner_kwargs.pop('rddl', None)
        planner_kwargs.pop('plan', None)
        planner_kwargs.pop('rollout_horizon', None)
        planner_kwargs.pop('optimizer_kwargs', None)
                        
        kwargs = {
            'rddl': self.env.model,
            'domain': self.env.domain_text,
            'instance': self.env.instance_text,
            'hyperparams_dict': self.hyperparams_dict,
            'timeout_training': self.timeout_training,
            'train_epochs': self.train_epochs,
            'planner_kwargs': planner_kwargs,
            'plan_kwargs': plan_kwargs,
            'verbose': self.verbose,
            'wrapped_bool_actions': self.wrapped_bool_actions,
            'eval_trials': self.eval_trials,
            'use_guess_last_epoch': self.use_guess_last_epoch
        }
        return objective_fn, kwargs


# ===============================================================================
# 
# DEEP REACTIVE POLICIES
#
# ===============================================================================
def objective_drp(params, kwargs, key, index):
                    
    # transform hyper-parameters to natural space
    param_values = [
        pmap(params[name])
        for (name, (*_, pmap)) in kwargs['hyperparams_dict'].items()
    ]
    
    # unpack hyper-parameters
    lr, w, layers, neurons = param_values                      
    if kwargs['verbose']:
        print(f'[{index}] key={key}, lr={lr}, w={w}, layers={layers}, neurons={neurons}...', flush=True)
           
    # initialize planning algorithm
    planner = JaxRDDLBackpropPlanner(
        rddl=deepcopy(kwargs['rddl']),
        plan=JaxDeepReactivePolicy(
            topology=[neurons] * layers,
            **kwargs['plan_kwargs']),
        optimizer_kwargs={'learning_rate': lr},
        **kwargs['planner_kwargs'])
    policy_hparams = {name: None for name in planner._action_bounds}
    model_params = {name: w for name in planner.compiled.model_params}
    
    # initialize policy
    key, subkey = jax.random.split(key)
    policy = JaxOfflineController(
        planner=planner,
        key=subkey,
        eval_hyperparams=policy_hparams,
        train_on_reset=True,
        epochs=kwargs['train_epochs'],
        train_seconds=kwargs['timeout_training'],
        model_params=model_params,
        policy_hyperparams=policy_hparams,
        verbose=0,
        tqdm_position=index)
    
    # initialize env for evaluation (need fresh copy to avoid concurrency)
    env = RDDLEnv(domain=kwargs['domain'],
                  instance=kwargs['instance'],
                  enforce_action_constraints=True)
    
    # perform training
    average_reward = 0.0
    for trial in range(kwargs['eval_trials']):
        key, subkey = jax.random.split(key)
        total_reward = policy.evaluate(env, seed=np.array(subkey)[0])['mean']
        if kwargs['verbose']:
            print(f'    [{index}] trial {trial + 1} key={subkey}, '
                  f'reward={total_reward}', flush=True)
        average_reward += total_reward / kwargs['eval_trials']
    if kwargs['verbose']:
        print(f'[{index}] average reward={average_reward}', flush=True)
    return average_reward


def power_two_int(x):
    return 2 ** int(x)


class JaxParameterTuningDRP(JaxParameterTuning):
    
    def __init__(self, *args,
                 hyperparams_dict: Dict[str, Tuple[float, float, Callable]]={
                    'lr': (-7., 2., power_ten),
                    'w': (0., 5., power_ten),
                    'layers': (1., 3., int),
                    'neurons': (2., 9., power_two_int)
                 },
                 **kwargs) -> None:
        '''Creates a new tuning class for deep reactive policies.
        
        :param *args: arguments to pass to parent class
        :param hyperparams_dict: same as parent class, but here must contain
        learning rate (lr), model weight (w), number of hidden layers (layers) 
        and number of neurons per hidden layer (neurons)
        :param **kwargs: keyword arguments to pass to parent class
        '''
        
        super(JaxParameterTuningDRP, self).__init__(
            *args, hyperparams_dict=hyperparams_dict, **kwargs)
    
    def _pickleable_objective_with_kwargs(self):
        objective_fn = objective_drp
        
        # duplicate planner and plan keyword arguments must be removed
        plan_kwargs = self.plan_kwargs.copy()
        plan_kwargs.pop('topology', None)
        
        planner_kwargs = self.planner_kwargs.copy()
        planner_kwargs.pop('rddl', None)
        planner_kwargs.pop('plan', None)
        planner_kwargs.pop('optimizer_kwargs', None)
                     
        kwargs = {
            'rddl': self.env.model,
            'domain': self.env.domain_text,
            'instance': self.env.instance_text,
            'hyperparams_dict': self.hyperparams_dict,
            'timeout_training': self.timeout_training,
            'train_epochs': self.train_epochs,
            'planner_kwargs': planner_kwargs,
            'plan_kwargs': plan_kwargs,
            'verbose': self.verbose,
            'eval_trials': self.eval_trials
        }
        return objective_fn, kwargs



from abc import ABCMeta, abstractmethod

class StateViz(metaclass=ABCMeta):
    @abstractmethod
    def render(self):
        pass
    
    # @abstractmethod
    # def build_nonfluents_layout(self):
    #     pass
    
    # @abstractmethod
    # def build_states_layout(self):
    #     pass

    # @abstractmethod
    # def init_canvas(self):
    #     pass
    
    # @abstractmethod
    # def build_object_layout(self):
    #     pass
    
    # @abstractmethod
    # def convert2img(self):
    #     pass
    




# This file is based on thiago pbueno's pyrddl.
# https://github.com/thiagopbueno/pyrddl
# it was adapted and extended for pyRDDLGym

from typing import Dict, List, Sequence, Tuple, Union

FluentTuple = Tuple[Union[str, None]]
Value = Union[bool, int, float]
FluentInitializer = Tuple[FluentTuple, Value]
FluentInitializerList = List[FluentInitializer]


class Instance(object):
    '''Instance class for accessing RDDL instance sections.
    Note:
        This class is intended to be solely used by the parser and compiler.
        Do not attempt to directly use this class to build an Instance object.
    Args:
        name: Name of RDDL instance.
        sections: Mapping from string to instance section.
    Attributes:
        name (str): Name of RDDL instance.
        domain (str): Name of RDDL domain.
        non_fluents (str): Name of RDDL non-fluents.
        init_state (:obj:`FluentInitializerList`): List of initial state initializers.
        max_nondef_actions (Union[int, str]): Maximum number of non-default actions.
        horizon (Union[int, str]): Number of decision timesteps.
        discount (float): Discount factor.
    '''

    def __init__(self, name: str, sections: Dict[str, Sequence]) -> None:
        self.name = name
        self.__dict__.update(sections)



import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt

import numpy as np
from PIL import Image

from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
from pyRDDLGym.Visualizer.StateViz import StateViz


class ColorVisualizer(StateViz):

    def __init__(self, model: PlanningModel,
                 steps_history=None,
                 figure_size=[10, 10],
                 dpi=100,
                 fontsize=10,
                 cmap='seismic',
                 loccol='black',
                 scale_full_history: bool=True,
                 ranges=None) -> None:
        self._model = model
        self._figure_size = figure_size
        self._dpi = dpi
        self._fontsize = fontsize
        self._cmap = cmap
        self._loccol = loccol
        self._scale_full_history = scale_full_history
        self._ranges = ranges
        
        self._fig, self._ax = None, None
        self._data = None
        self._img = None
        
        if steps_history is None:
            steps_history = model.horizon
        self._steps = steps_history
        self._state_hist = {}
        self._state_shapes = {}
        self._labels = {}
        self._historic_min_max = {}
        for (state, values) in self._model.states.items():
            values = np.atleast_1d(values)
            self._state_hist[state] = np.full(shape=(len(values), self._steps),
                                              fill_value=np.nan)
            self._state_shapes[state] = self._model.object_counts(
                self._model.param_types[state])
            self._labels[state] = list(map(
                ','.join, self._model.variations(self._model.param_types[state])
            ))
            self._historic_min_max[state] = (np.inf, -np.inf)
        self._step = 0
        
    def convert2img(self, fig, ax): 
        # ax.set_position((0, 0, 1, 1))
        fig.canvas.draw()

        data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        img = Image.fromarray(data)

        self._data = data
        self._img = img
        return img

    def render(self, state):
        
        # update the state info
        if self._step >= self._steps:
            for (_, values) in self._state_hist.items():
                values[:,:-1] = values[:, 1:]
        states = {name: np.full(shape=shape, fill_value=np.nan)
                  for (name, shape) in self._state_shapes.items()}
        for (name, value) in state.items():
            var, objects = self._model.parse(name)
            states[var][self._model.indices(objects)] = value
        index = min(self._step, self._steps - 1)
        for (name, values) in states.items():
            self._state_hist[name][:, index] = np.ravel(values, order='C')
            
        # draw color plots
        self._fig, self._ax = plt.subplots(len(self._state_hist), 1,
                                           squeeze=True,
                                           figsize=self._figure_size,
                                           sharex=True) 
        if len(self._state_hist) == 1:
            self._ax = (self._ax,)
            
        for (y, (state, values)) in enumerate(self._state_hist.items()):
            values = values[::-1, :]
            
            # title, labels and cursor line
            self._ax[y].xaxis.label.set_fontsize(self._fontsize)
            self._ax[y].yaxis.label.set_fontsize(self._fontsize)
            self._ax[y].title.set_text(state)
            self._ax[y].set_xlabel('decision epoch')
            self._ax[y].set_ylabel(state)
            self._ax[y].axvline(x=index + 0.5, 
                                ymin=0.0, 
                                ymax=1.0,
                                color=self._loccol, 
                                linestyle='--', 
                                linewidth=2, 
                                alpha=0.9)
            
            # scaling of y axis
            vmin, vmax = None, None
            if self._scale_full_history:
                hmin, hmax = self._historic_min_max[state]
                valid_values = values[np.isfinite(values)]
                vmin = min(hmin, np.min(valid_values))
                vmax = max(hmax, np.max(valid_values))
                self._historic_min_max[state] = (vmin, vmax)            
            if self._ranges is not None and state in self._ranges:
                vmin, vmax = self._ranges[state]    
            
            # plot fluent as color mesh
            im = self._ax[y].pcolormesh(values, 
                                        edgecolors=self._loccol, 
                                        linewidth=0.5, 
                                        cmap=self._cmap, 
                                        vmin=vmin, 
                                        vmax=vmax)         
            plt.colorbar(im, ax=self._ax[y])
            labels = self._labels[state]
            self._ax[y].yaxis.set_ticks([0, len(labels)])
            self._ax[y].yaxis.set(ticks=np.arange(0.5, len(labels), 1), ticklabels=labels) 
            self._ax[y].set_yticklabels(labels,
                                        fontdict={"fontsize": self._fontsize},
                                        rotation=30)
            
        self._step = self._step + 1
        plt.tight_layout()
        
        img = self.convert2img(self._fig, self._ax)
        
        plt.clf()
        plt.close()

        return img
    



from typing import Dict, List, Union

from pyRDDLGym.Core.Parser.expr import Expression
from pyRDDLGym.Core.Parser.cpf import CPF


class RDDLDecompiler:
    '''Converts AST representation (e.g., Expression) to a string that represents
    the corresponding expression in RDDL.'''
    
    def decompile_expr(self, expr: Expression, level: int=0) -> str:
        '''Converts an AST expression to a string representing valid RDDL code.
        
        :param expr: the expression to convert
        :param level: indentation level
        '''
        return self._decompile(expr, False, level)
    
    def decompile_cpf(self, cpf: CPF, level: int=0) -> str:
        '''Converts a CPF object to its equivalent string as it would appear in
        the cpfs {...} block in a RDDL domain description. 
        
        :param cpf: the CPF object to convert
        :param level: indentation level
        '''
        lhs = self._symbolic(*cpf.pvar[1], aggregation=False)
        rhs = self.decompile_expr(cpf.expr, level)
        return f'{lhs} = {rhs};'
    
    def decompile_exprs(self, rddl, level: int=0) -> Dict[str, Union[str, Dict[str, str], List[str]]]:
        '''Converts a RDDL model to a dictionary of decompiled expression strings,
        as they would appear in the domain description file.'''
        decompiled = {}
        decompiled['cpfs'] = {name: self.decompile_expr(expr, level)
                              for (name, (_, expr)) in rddl.cpfs.items()}
        decompiled['reward'] = self.decompile_expr(rddl.reward, level)
        decompiled['invariants'] = [self.decompile_expr(expr, level) 
                                    for expr in rddl.invariants]
        decompiled['preconditions'] = [self.decompile_expr(expr, level) 
                                       for expr in rddl.preconditions]
        decompiled['terminations'] = [self.decompile_expr(expr, level) 
                                      for expr in rddl.terminals]
        return decompiled
    
    def decompile_domain(self, rddl) -> str:
        '''Converts a RDDL model to a RDDL domain description file.'''
        decompiled = self.decompile_exprs(rddl, level=2)
        
        decompiled_types = ''
        if rddl.objects:
            decompiled_types = '\n\ttypes {'
            for (name, values) in rddl.objects.items():
                if name in rddl.enums:
                    decompiled_types += f'\n\t\t{name}: {{ ' + \
                        ', '.join([f'@{v}' for v in values]) + ' };'
                else:
                    decompiled_types += f'\n\t\t{name}: object;'
            decompiled_types += '\n\t};'
        
        decompiled_pvars = '\n\tpvariables {'
        for pvars in [rddl.nonfluents, rddl.derived, rddl.interm, 
                      rddl.states, rddl.observ, rddl.actions]:
            if pvars:
                decompiled_pvars += '\n'
            for name in pvars:
                prange = rddl.variable_ranges[name]
                ptype = rddl.variable_types[name]
                decompiled_params = ''
                if rddl.param_types[name]:
                    decompiled_params = '(' + ', '.join(rddl.param_types[name]) + ')' 
                dv = str(rddl.default_values[name])
                if dv == 'True' or dv == 'False':
                    dv = dv.lower()
                if prange not in ['real', 'int', 'bool']:
                    dv = f'@{dv}'
                if ptype in ['interm-fluent', 'derived-fluent', 'observ-fluent']:
                    decompiled_pvars += f'\n\t\t{name}{decompiled_params} : ' + \
                                        f'{{ {ptype}, {prange} }};'
                else:
                    decompiled_pvars += f'\n\t\t{name}{decompiled_params} : ' + \
                                        f'{{ {ptype}, {prange}, default = {dv} }};'
        decompiled_pvars += '\n\t};'
        
        decompiled_cpfs = '\n\tcpfs {'
        for (name, expr) in decompiled['cpfs'].items():
            decompiled_cpfs += f'\n\n\t\t{name} = {expr};'
        decompiled_cpfs += '\n\t};'
        
        decompiled_reward = f'\n\treward = {decompiled["reward"]};'
        
        decompiled_invariants = ''
        if decompiled['invariants']:
            decompiled_invariants = '\n\n\tstate-invariants {'
            for expr in decompiled['invariants']:
                decompiled_invariants += f'\n\t\t{expr};'
            decompiled_invariants += '\n\t};'
        
        decompiled_preconds = ''
        if decompiled['preconditions']:
            decompiled_preconds = '\n\n\taction-preconditions {'
            for expr in decompiled['preconditions']:
                decompiled_preconds += f'\n\t\t{expr};'
            decompiled_preconds += '\n\t};'
        
        decompiled_terminals = ''
        if decompiled['terminations']:
            decompiled_terminals = '\n\n\ttermination {'
            for expr in decompiled['terminations']:
                decompiled_terminals += f'\n\t\t{expr};'
            decompiled_terminals += '\n\t};'
        
        decompiled_domain = (
            f'domain {rddl.domainName()} {{'
            f'\n{decompiled_types}'
            f'\n{decompiled_pvars}'
            f'\n{decompiled_cpfs}'
            f'\n{decompiled_reward}'
            f'{decompiled_invariants}'
            f'{decompiled_preconds}'
            f'{decompiled_terminals}'
            f'\n}}')
        return decompiled_domain
        
    def _decompile(self, expr, enclose, level):
        etype, _ = expr.etype
        if etype == 'constant':
            return self._decompile_constant(expr, enclose, level)
        elif etype == 'pvar':
            return self._decompile_pvar(expr, enclose, level)
        elif etype in {'arithmetic', 'relational', 'boolean'}:
            return self._decompile_math(expr, enclose, level)
        elif etype == 'aggregation':
            return self._decompile_aggregation(expr, enclose, level)   
        elif etype == 'func':
            return self._decompile_func(expr, enclose, level) 
        elif etype == 'control':
            return self._decompile_control(expr, enclose, level)        
        elif etype == 'randomvar':
            return self._decompile_random(expr, enclose, level)
        elif etype == 'randomvector':
            return self._decompile_random_vector(expr, enclose, level)
        elif etype == 'matrix':
            return self._decompile_matrix(expr, enclose, level)
        else:
            return ''
    
    def _symbolic(self, value, params, aggregation):
        value = str(value)
        if params is not None and params:
            if aggregation:
                args = ', '.join(f'{k}: {v}' for (k, v) in params)
                value += f'_{{{args}}}'
            else:
                args = ', '.join(map(str, params))
                value += f'({args})'
        return value

    def _decompile_constant(self, expr, enclose, level):
        value = expr.args
        value = str(value)
        if value == 'True':
            value = 'true'
        elif value == 'False':
            value = 'false'
        return value
        
    def _decompile_pvar(self, expr, enclose, level):
        _, name = expr.etype
        _, params = expr.args 
        if params is not None:
            params = ((self._decompile(arg, False, 0) 
                       if isinstance(arg, Expression) 
                       else arg)
                      for arg in params)
        return self._symbolic(name, params, aggregation=False)
        
    def _decompile_math(self, expr, enclose, level):
        _, op = expr.etype
        args = expr.args
        if len(args) == 1:
            arg, = args
            value = str(op) + self._decompile(arg, True, level)
        else:
            sep = ' ' + str(op) + ' '
            value = sep.join(self._decompile(arg, True, level) for arg in args)
            
        if enclose:
            value = f'( {value} )'
        return value
        
    def _decompile_aggregation(self, expr, enclose, level):
        _, op = expr.etype
        * pvars, arg = expr.args
        params = [pvar for (_, pvar) in pvars]
        agg = self._symbolic(op, params, aggregation=True)
        decompiled = self._decompile(arg, False, level)        
        return f'( {agg} [ {decompiled} ] )'
        
    def _decompile_func(self, expr, enclose, level):
        _, op = expr.etype
        decompiled = ', '.join(self._decompile(arg, False, level) 
                               for arg in expr.args)
        return f'{op}[{decompiled}]'
            
    def _decompile_control(self, expr, enclose, level):
        _, op = expr.etype
        indent = '\t' * (level + 1)
        
        if op == 'if':
            pred, if_true, if_false = expr.args
            pred = self._decompile(pred, False, level)
            if_true = self._decompile(if_true, True, level + 1)
            if_false = self._decompile(if_false, True, level + 1)
            value = f'if ({pred})\n{indent}then {if_true}\n{indent}else {if_false}'

        else:  # switch
            pvar, *args = expr.args
            pred = self._decompile(pvar, False, level)
            cases = [''] * len(args)
            for (i, _case) in enumerate(args):
                case_type, value = _case
                if case_type == 'case':
                    literal, arg = value
                    decompiled = self._decompile(arg, False, level + 1)
                    cases[i] = f'case {literal} : {decompiled}'
                else:  # default
                    decompiled = self._decompile(value, False, level + 1)
                    cases[i] = f'default : {decompiled}'
            cases = f',\n{indent}'.join(cases)
            value = f'switch({pred}) {{ \n{indent}{cases}\n }}'

        if enclose: 
            value = f'( {value} )'
        return value
    
    def _decompile_random(self, expr, enclose, level):
        _, op = expr.etype
        
        if op == 'Discrete' or op == 'UnnormDiscrete':
            (_, var), *args = expr.args
            cases = [var] + [''] * len(args)
            for (i, _case) in enumerate(args):
                _, (literal, arg) = _case
                decompiled = self._decompile(arg, False, level + 1)
                cases[i + 1] = f'{literal} : {decompiled}'
            indent = '\t' * (level + 1)
            value = f',\n{indent}'.join(cases)
        
        elif op == 'Discrete(p)' or op == 'UnnormDiscrete(p)':
            op = op[:-3]
            * pvars, args = expr.args
            params = [pvar for (_, pvar) in pvars]
            op = self._symbolic(op, params, aggregation=True)    
            value = ', '.join(self._decompile(arg, False, level) 
                              for arg in args)
            
        else:  # Normal, exponential, etc...
            value = ', '.join(self._decompile(arg, False, level) 
                              for arg in expr.args)
            
        return f'{op}({value})'
    
    def _decompile_random_vector(self, expr, enclose, level):
        _, op = expr.etype
        pvars, args = expr.args
        args = ', '.join(self._decompile(arg, False, level) for arg in args)
        pvars = ', '.join(pvars)
        return f'{op}[{pvars}]({args})'
            
    def _decompile_matrix(self, expr, enclose, level):
        _, op = expr.etype
        if op == 'det':
            *pvars, arg = expr.args
            params = [pvar for (_, pvar) in pvars]
            agg = self._symbolic(op, params, aggregation=True)
            decompiled = self._decompile(arg, False, level)        
            return f'{agg}[ {decompiled} ]'
        elif op == 'inverse' or op == 'pinverse' or op == 'cholesky':
            pvars, arg = expr.args
            prow, pcol = pvars
            params = f'row={prow}, col={pcol}'
            decompiled = self._decompile(arg, False, level)
            return f'{op}[{params}][ {decompiled} ]'



from pyRDDLGym import ExampleManager
from pyRDDLGym.Core.Parser.RDDLReader import RDDLReader
from pyRDDLGym.Core.Parser.parser import RDDLParser, RDDLlex

ENV = 'PropDBN'

def main():
    # get the environment info
    EnvInfo = ExampleManager.GetEnvInfo(ENV)

    # read and parse domain and instance
    reader = RDDLReader(EnvInfo.get_domain(), EnvInfo.get_instance(0))
    domain = reader.rddltxt

    # parse RDDL file
    MyLexer = RDDLlex()
    MyLexer.build()
    MyLexer.input(domain)
    a = [token for token in MyLexer._lexer]

    parser = RDDLParser(lexer=None, verbose=False)
    parser.build()
    rddl = parser.parse(domain)
    print(rddl)






if __name__ == "__main__":
    main()



import sys

from pyRDDLGym import ExampleManager
from pyRDDLGym.Core.Policies.RDDLSimAgent import RDDLSimAgent

def main(domain):
    EnvInfo = ExampleManager.GetEnvInfo(domain)
    agent = RDDLSimAgent(EnvInfo.get_domain(), EnvInfo.get_instance(1), 30, 300)
    agent.run()

if __name__ == "__main__":
    domain = "recon2018"
    if len(sys.argv) == 2:
        domain = sys.argv[1]
    main(domain)



import glob
import os
from PIL import Image
import warnings
import re

# (mike: #166) opencv is now optional
try:
    import cv2
    _ALLOW_MP4 = True
except:
    warnings.warn('cv2 is not installed: save_as_mp4 option will be disabled.',
                  stacklevel=2)
    _ALLOW_MP4 = False


class MovieGenerator:
    
    def __init__(self,
                 save_dir: str,
                 env_name: str,
                 max_frames: int,
                 skip: int=1,
                 save_format: str='png',
                 frame_duration: int=100,
                 loop: int=0,
                 save_as_mp4: bool=False):
        '''Creates a new movie generator for saving frames to disk, and creating animated GIFs.
        
        :param save_dir: the directory to save images to
        :param env_name: the root name of each image file
        :param max_frames: the max number of frames to save
        :param skip: how often frames should be recorded
        :param save_format: the format in which to save individual frames
        :param frame_duration: the duration of each frame in the animated video
        :param loop: how many times the animated GIF should loop
        :param save_as_mp4: whether to save mp4 video (or GIF if False)
        '''
        self.save_dir = save_dir
        self.save_path = os.path.join(save_dir, env_name + '_{}' + '.' + save_format)
        self.env_name = env_name
        self.max_frames = max_frames
        self.skip = skip
        self.frame_duration = frame_duration
        self.loop = loop
        self.save_as_mp4 = save_as_mp4
        
        self._n_frame = 0
        self._time = 0
    
    def reset(self) -> None:
        load_path = self.save_path.format('*')
        files = glob.glob(load_path)
        removed = 0
        for file in files:
            os.remove(file)
            removed += 1
        if removed:
            warnings.warn(f'removed {removed} temporary files at {load_path}',
                          stacklevel=2)
        
        self._n_frame = 0
        self._time = 0
        
    def save_frame(self, image) -> None:
        if self._n_frame >= self.max_frames:
            return     
        if self._time % self.skip != 0: 
            self._time += 1
            return
        
        file_path = self.save_path.format(
            str(self._n_frame).rjust(10, '0'))
        image.save(file_path)
        self._n_frame += 1 
        self._time += 1
    
    def save_animation(self, file_name: str=None):
        if _ALLOW_MP4 and self.save_as_mp4:
            self.save_mp4(file_name)
        else:
            self.save_gif(file_name)
        self.reset()
            
    def save_gif(self, file_name: str=None):
        if file_name is None:
            file_name = self.env_name
        load_path = self.save_path.format('*')

        def getOrder(frame):
            return int(re.search('\d+', frame).group(0))

        files = glob.glob(load_path)
        files.sort(key=getOrder)

        # images = map(Image.open, glob.glob(load_path))
        images = map(Image.open, files)
        
        save_path = os.path.join(self.save_dir, file_name + '.gif')
        frame0 = next(images, None)
        if frame0 is not None:
            frame0.save(fp=save_path,
                        format='GIF',
                        append_images=images,
                        save_all=True,
                        duration=self.frame_duration,
                        loop=self.loop)  
        
    def save_mp4(self, file_name: str=None):
        if file_name is None:
            file_name = self.env_name
        load_path = self.save_path.format('*')
        images = glob.glob(load_path)
        
        writer, w, h = None, None, None
        fps = 1000 / self.frame_duration
        for file in images:
            frame = cv2.imread(file)
            if w is None:
                h, w, _ = frame.shape
                fourcc = cv2.VideoWriter_fourcc('m', 'p', '4', 'v')
                writer = cv2.VideoWriter(file_name + '.mp4', fourcc, fps, (w, h))
            writer.write(frame)
        if writer is not None:
            writer.release()



import numpy as np
np.seterr(all='raise')
from typing import Dict, Set, Union

from pyRDDLGym.Core.ErrorHandling.RDDLException import print_stack_trace
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLActionPreconditionNotSatisfiedError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidActionError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLNotImplementedError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLStateInvariantNotSatisfiedError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLUndefinedVariableError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLValueOutOfRangeError

from pyRDDLGym.Core.Compiler.RDDLLevelAnalysis import RDDLLevelAnalysis
from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
from pyRDDLGym.Core.Compiler.RDDLObjectsTracer import RDDLObjectsTracer
from pyRDDLGym.Core.Compiler.RDDLValueInitializer import RDDLValueInitializer
from pyRDDLGym.Core.Debug.Logger import Logger
from pyRDDLGym.Core.Parser.expr import Value

Args = Dict[str, Value]

        
class RDDLSimulator:
    
    def __init__(self, rddl: PlanningModel,
                 allow_synchronous_state: bool=True,
                 rng: np.random.Generator=np.random.default_rng(),
                 logger: Logger=None,
                 keep_tensors: bool=False) -> None:
        '''Creates a new simulator for the given RDDL model.
        
        :param rddl: the RDDL model
        :param allow_synchronous_state: whether state-fluent can be synchronous
        :param rng: the random number generator
        :param logger: to log information about compilation to file
        :param keep_tensors: whether the sampler takes actions and
        returns state in numpy array form
        '''
        self.rddl = rddl
        self.allow_synchronous_state = allow_synchronous_state
        self.rng = rng
        self.logger = logger
        self.keep_tensors = keep_tensors
        
        self._compile()
        
        # basic operations
        self.ARITHMETIC_OPS = {
            '+': np.add,
            '-': np.subtract,
            '*': np.multiply,
            '/': np.divide
        }    
        self.RELATIONAL_OPS = {
            '>=': np.greater_equal,
            '<=': np.less_equal,
            '<': np.less,
            '>': np.greater,
            '==': np.equal,
            '~=': np.not_equal
        }
        self.LOGICAL_OPS = {
            '^': np.logical_and,
            '&': np.logical_and,
            '|': np.logical_or,
            '~': np.logical_xor,
            '=>': lambda x, y: np.logical_or(np.logical_not(x), y),
            '<=>': np.equal
        }
        self.AGGREGATION_OPS = {
            'sum': np.sum,
            'avg': np.mean,
            'prod': np.prod,
            'minimum': np.min,
            'maximum': np.max,
            'forall': np.all,
            'exists': np.any,
            'argmin': np.argmin,
            'argmax': np.argmax
        }
        self.AGGREGATION_BOOL = {'forall', 'exists'}
        self.UNARY = {        
            'abs': np.abs,
            'sgn': lambda x: np.sign(x).astype(RDDLValueInitializer.INT),
            'round': lambda x: np.round(x).astype(RDDLValueInitializer.INT),
            'floor': lambda x: np.floor(x).astype(RDDLValueInitializer.INT),
            'ceil': lambda x: np.ceil(x).astype(RDDLValueInitializer.INT),
            'cos': np.cos,
            'sin': np.sin,
            'tan': np.tan,
            'acos': np.arccos,
            'asin': np.arcsin,
            'atan': np.arctan,
            'cosh': np.cosh,
            'sinh': np.sinh,
            'tanh': np.tanh,
            'exp': np.exp,
            'ln': np.log,
            'sqrt': np.sqrt,
            'lngamma': lngamma,
            'gamma': lambda x: np.exp(lngamma(x))
        }        
        self.BINARY = {
            'div': lambda x, y: np.floor_divide(x, y).astype(RDDLValueInitializer.INT),
            'mod': lambda x, y: np.mod(x, y).astype(RDDLValueInitializer.INT),
            'fmod': lambda x, y: np.mod(x, y),
            'min': np.minimum,
            'max': np.maximum,
            'pow': np.power,
            'log': lambda x, y: np.log(x) / np.log(y),
            'hypot': np.hypot
        }
        self.BINARY_REQUIRES_INT = {'div', 'mod'}
        self.CONTROL_OPS = {'if': np.where,
                            'switch': np.select}
    
    def seed(self, seed: int) -> None:
        '''Sets the pseudo-random RNG seed for generating random numbers.
        
        :param seed: seed value to use for all future simulations
        '''
        self.rng = np.random.default_rng(seed)
        
    def _compile(self):
        rddl = self.rddl
        
        # compile initial values
        if self.logger is not None:
            self.logger.clear()
        initializer = RDDLValueInitializer(rddl, logger=self.logger)
        self.init_values = initializer.initialize()
        
        # compute dependency graph for CPFs and sort them by evaluation order
        sorter = RDDLLevelAnalysis(rddl, 
                                   allow_synchronous_state=self.allow_synchronous_state, 
                                   logger=self.logger)
        levels = sorter.compute_levels()      
        self.cpfs = []  
        for cpfs in levels.values():
            for cpf in cpfs:
                _, expr = rddl.cpfs[cpf]
                prange = rddl.variable_ranges[cpf]
                dtype = RDDLValueInitializer.NUMPY_TYPES.get(
                    prange, RDDLValueInitializer.INT)
                self.cpfs.append((cpf, expr, dtype))
                
        # trace expressions to cache information to be used later
        tracer = RDDLObjectsTracer(rddl, logger=self.logger, cpf_levels=levels)
        self.traced = tracer.trace()
        
        # initialize all fluent and non-fluent values        
        self.subs = self.init_values.copy()
        self.state = None  
        self.noop_actions = {var: values
                             for (var, values) in self.init_values.items()
                             if rddl.variable_types[var] == 'action-fluent'}
        self._pomdp = bool(rddl.observ)
        
        # cached for performance
        self.invariant_names = [f'Invariant {i}' for i in range(len(rddl.invariants))]        
        self.precond_names = [f'Precondition {i}' for i in range(len(rddl.preconditions))]
        self.terminal_names = [f'Termination {i}' for i in range(len(rddl.terminals))]
        
        self.grounded_actionsranges = rddl.groundactionsranges()
        self.grounded_noop_actions = rddl.ground_values_from_dict(self.noop_actions)
        
    @property
    def states(self) -> Args:
        return self.state.copy()

    @property
    def isPOMDP(self) -> bool:
        return self._pomdp
    
    # ===========================================================================
    # error checks
    # ===========================================================================
    
    @staticmethod
    def _check_type(value, valid, msg, expr, arg=None):
        if not np.can_cast(value, valid):
            dtype = getattr(value, 'dtype', type(value))
            if arg is None:
                raise RDDLTypeError(
                    f'{msg} must evaluate to {valid}, '
                    f'got {value} of type {dtype}.\n' + print_stack_trace(expr))
            else:
                raise RDDLTypeError(
                    f'Argument {arg} of {msg} must evaluate to {valid}, '
                    f'got {value} of type {dtype}.\n' + print_stack_trace(expr))
    
    @staticmethod
    def _check_types(value, valid, msg, expr):
        for valid_type in valid:
            if np.can_cast(value, valid_type):
                return
        dtype = getattr(value, 'dtype', type(value))
        raise RDDLTypeError(
            f'{msg} must evaluate to one of {valid}, '
            f'got {value} of type {dtype}.\n' + print_stack_trace(expr))
    
    @staticmethod
    def _check_op(op, valid, msg, expr):
        numpy_op = valid.get(op, None)
        if numpy_op is None:
            raise RDDLNotImplementedError(
                f'{msg} operator {op} is not supported: '
                f'must be in {set(valid.keys())}.\n' + print_stack_trace(expr))
        return numpy_op
        
    @staticmethod
    def _check_arity(args, required, msg, expr):
        if len(args) != required:
            raise RDDLInvalidNumberOfArgumentsError(
                f'{msg} requires {required} argument(s), got {len(args)}.\n' + 
                print_stack_trace(expr))
    
    @staticmethod
    def _check_positive(value, strict, msg, expr):
        if strict and not np.all(value > 0):
            raise RDDLValueOutOfRangeError(
                f'{msg} must be positive, got {value}.\n' + 
                print_stack_trace(expr))
        elif not strict and not np.all(value >= 0):
            raise RDDLValueOutOfRangeError(
                f'{msg} must be non-negative, got {value}.\n' + 
                print_stack_trace(expr))
    
    @staticmethod
    def _check_bounds(lb, ub, msg, expr):
        if not np.all(lb <= ub):
            raise RDDLValueOutOfRangeError(
                f'Bounds of {msg} are invalid:' 
                f'max value {ub} must be >= min value {lb}.\n' + 
                print_stack_trace(expr))
            
    @staticmethod
    def _check_range(value, lb, ub, msg, expr):
        if not np.all(np.logical_and(value >= lb, value <= ub)):
            raise RDDLValueOutOfRangeError(
                f'{msg} must be in the range [{lb}, {ub}], got {value}.\n' + 
                print_stack_trace(expr))
    
    # ===========================================================================
    # main sampling routines
    # ===========================================================================
    
    def _process_actions(self, actions):
        rddl = self.rddl
        
        # override new_actions with any new actions
        if self.keep_tensors:
            new_actions = self.noop_actions.copy()
            for (action, value) in actions.items(): 
                if action in new_actions:
                    new_actions[action] = value
        else:            
            new_actions = {action: np.copy(value) 
                           for (action, value) in self.noop_actions.items()}
            for (action, value) in actions.items(): 
                value = rddl.index_of_object.get(value, value)
                if action in new_actions:
                    new_actions[action] = value
                else:
                    var, objects = rddl.parse(action)
                    tensor = new_actions.get(var, None)
                    if tensor is None: 
                        raise RDDLInvalidActionError(
                            f'<{action}> is not a valid action-fluent, ' 
                            f'must be one of {set(new_actions.keys())}.')
                    RDDLSimulator._check_type(value, tensor.dtype, action, expr='')            
                    tensor[rddl.indices(objects)] = value
                
        return new_actions
    
    def check_state_invariants(self, silent: bool=False) -> bool:
        '''Throws an exception if the state invariants are not satisfied.'''
        for (i, invariant) in enumerate(self.rddl.invariants):
            loc = self.invariant_names[i]
            sample = self._sample(invariant, self.subs)
            RDDLSimulator._check_type(sample, bool, loc, invariant)
            if not bool(sample):
                if not silent:
                    raise RDDLStateInvariantNotSatisfiedError(
                        f'{loc} is not satisfied.\n' + print_stack_trace(invariant))
                return False
        return True
    
    def check_default_action_count(self, actions: Args, 
                                   enforce_for_non_bool: bool=True) -> None:
        '''Throws an exception if the actions do not satisfy max-nondef-actions.'''     
        if self.keep_tensors:
            action_ranges = self.rddl.actionsranges
            noop_actions = self.noop_actions
        else:
            action_ranges = self.grounded_actionsranges
            noop_actions = self.grounded_noop_actions
            
        total_non_default = 0
        for (var, values) in actions.items():
            
            # check that action is valid
            prange = action_ranges.get(var, None)
            if prange is None:
                raise RDDLInvalidActionError(
                    f'<{var}> is not a valid action fluent, '
                    f'must be one of {set(action_ranges.keys())}.')
            
            # check that action shape is valid
            default_values = noop_actions[var]
            if np.shape(values) != np.shape(default_values):
                raise RDDLInvalidActionError(
                    f'Value array for action <{var}> must be of shape '
                    f'{np.shape(default_values)}, got array of shape '
                    f'{np.shape(values)}.')
            
            # accumulate count of non-default actions
            if enforce_for_non_bool or prange == 'bool':
                total_non_default += np.count_nonzero(values != default_values)
        
        if total_non_default > self.rddl.max_allowed_actions:
            raise RDDLInvalidActionError(
                f'Expected at most {self.rddl.max_allowed_actions} '
                f'non-default actions, got {total_non_default}.')
        
    def check_action_preconditions(self, actions: Args, silent: bool=False) -> bool:
        '''Throws an exception if the action preconditions are not satisfied.'''     
        actions = self._process_actions(actions)
        self.subs.update(actions)
        
        for (i, precond) in enumerate(self.rddl.preconditions):
            loc = self.precond_names[i]
            sample = self._sample(precond, self.subs)
            RDDLSimulator._check_type(sample, bool, loc, precond)
            if not bool(sample):
                if not silent:
                    raise RDDLActionPreconditionNotSatisfiedError(
                        f'{loc} is not satisfied for actions {actions}.\n' + 
                        print_stack_trace(precond))
                return False
        return True
    
    def check_terminal_states(self) -> bool:
        '''Return True if a terminal state has been reached.'''
        for (i, terminal) in enumerate(self.rddl.terminals):
            loc = self.terminal_names[i]
            sample = self._sample(terminal, self.subs)
            RDDLSimulator._check_type(sample, bool, loc, terminal)
            if bool(sample):
                return True
        return False
    
    def sample_reward(self) -> float:
        '''Samples the current reward given the current state and action.'''
        return float(self._sample(self.rddl.reward, self.subs))
    
    def reset(self) -> Union[Dict[str, None], Args]:
        '''Resets the state variables to their initial values.'''
        rddl = self.rddl
        subs = self.subs = self.init_values.copy()
        keep_tensors = self.keep_tensors
        
        # update state
        self.state = {}
        for state in rddl.states:
            if keep_tensors:
                self.state[state] = subs[state]
            else:
                self.state.update(rddl.ground_values(state, subs[state]))
        
        # update observation
        if self._pomdp:
            obs = {var: None for var in rddl.observ}
        else:
            obs = self.state
        
        done = self.check_terminal_states()
        return obs, done
    
    def step(self, actions: Args) -> Args:
        '''Samples and returns the next state from the CPF expressions.
        
        :param actions: a dict mapping current action fluent to their values
        '''
        rddl = self.rddl
        keep_tensors = self.keep_tensors
        actions = self._process_actions(actions)
        subs = self.subs
        subs.update(actions)
        
        # evaluate CPFs in topological order
        for (cpf, expr, dtype) in self.cpfs:
            sample = self._sample(expr, subs)
            RDDLSimulator._check_type(sample, dtype, cpf, expr)
            subs[cpf] = sample
        
        # evaluate reward
        reward = self.sample_reward()
        
        # update state
        self.state = {}
        for (state, next_state) in rddl.next_state.items():
            subs[state] = subs[next_state]
            if keep_tensors:
                self.state[state] = subs[state]
            else:
                self.state.update(rddl.ground_values(state, subs[state]))
        
        # update observation
        if self._pomdp: 
            obs = {}
            for var in rddl.observ:
                if keep_tensors:
                    obs[var] = subs[var]
                else:
                    obs.update(rddl.ground_values(var, subs[var]))
        else:
            obs = self.state
        
        done = self.check_terminal_states()        
        return obs, reward, done
        
    # ===========================================================================
    # start of sampling subroutines
    # ===========================================================================
    
    def _sample(self, expr, subs):
        etype, _ = expr.etype
        if etype == 'constant':
            return self._sample_constant(expr, subs)
        elif etype == 'pvar':
            return self._sample_pvar(expr, subs)
        elif etype == 'arithmetic':
            return self._sample_arithmetic(expr, subs)
        elif etype == 'relational':
            return self._sample_relational(expr, subs)
        elif etype == 'boolean':
            return self._sample_logical(expr, subs)
        elif etype == 'aggregation':
            return self._sample_aggregation(expr, subs)
        elif etype == 'func':
            return self._sample_func(expr, subs)
        elif etype == 'control':
            return self._sample_control(expr, subs)
        elif etype == 'randomvar':
            return self._sample_random(expr, subs)
        elif etype == 'randomvector':
            return self._sample_random_vector(expr, subs)
        elif etype == 'matrix':
            return self._sample_matrix(expr, subs)
        else:
            raise RDDLNotImplementedError(
                f'Internal error: expression type {etype} is not supported.\n' + 
                print_stack_trace(expr))
                
    # ===========================================================================
    # leaves
    # ===========================================================================
        
    def _sample_constant(self, expr, _):
        return self.traced.cached_sim_info(expr)
    
    def _sample_pvar(self, expr, subs):
        var, args = expr.args
        
        # free variable (e.g., ?x) and object converted to canonical index
        is_value, cached_info = self.traced.cached_sim_info(expr)
        if is_value:
            return cached_info
        
        # extract variable value
        sample = subs.get(var, None)
        if sample is None:
            raise RDDLUndefinedVariableError(
                f'Variable <{var}> is referenced before assignment.\n' + 
                print_stack_trace(expr))
        
        # lifted domain must slice and/or reshape value tensor
        if cached_info is not None:
            slices, axis, shape, op_code, op_args = cached_info
            if slices: 
                if op_code == RDDLObjectsTracer.NUMPY_OP_CODE.NESTED_SLICE:
                    slices = tuple(
                        (self._sample(arg, subs) if _slice is None else _slice)
                        for (arg, _slice) in zip(args, slices)
                    )
                sample = sample[slices]
            if axis:
                sample = np.expand_dims(sample, axis=axis)
                sample = np.broadcast_to(sample, shape=shape)
            if op_code == RDDLObjectsTracer.NUMPY_OP_CODE.EINSUM:
                sample = np.einsum(sample, *op_args)
            elif op_code == RDDLObjectsTracer.NUMPY_OP_CODE.TRANSPOSE:
                sample = np.transpose(sample, axes=op_args)
        return sample
    
    # ===========================================================================
    # arithmetic
    # ===========================================================================
    
    def _sample_arithmetic(self, expr, subs):
        _, op = expr.etype
        numpy_op = RDDLSimulator._check_op(
            op, self.ARITHMETIC_OPS, 'Arithmetic', expr)
        
        args = expr.args        
        n = len(args)
        
        # unary negation
        if n == 1 and op == '-':
            arg, = args
            return -1 * self._sample(arg, subs)
        
        # binary operator: for * try to short-circuit if possible
        elif n == 2:
            lhs, rhs = args
            if op == '*':
                return self._sample_product(lhs, rhs, subs)
            else:
                sample_lhs = 1 * self._sample(lhs, subs)
                sample_rhs = 1 * self._sample(rhs, subs)
                try:
                    return numpy_op(sample_lhs, sample_rhs)
                except:
                    raise ArithmeticError(
                        f'Cannot evaluate arithmetic operation {op} '
                        f'at {sample_lhs} and {sample_rhs}.\n' + 
                        print_stack_trace(expr))
        
        # for a grounded domain can short-circuit * and +
        elif n > 0 and not self.traced.cached_objects_in_scope(expr):
            if op == '*':
                return self._sample_product_grounded(args, subs)
            elif op == '+':
                return sum(1 * self._sample(arg, subs) for arg in args)
        
        raise RDDLInvalidNumberOfArgumentsError(
            f'Arithmetic operator {op} does not have the required '
            f'number of arguments.\n' + print_stack_trace(expr))
    
    def _sample_product(self, lhs, rhs, subs):
        
        # prioritize simple expressions
        if rhs.is_constant_expression() or rhs.is_pvariable_expression():
            lhs, rhs = rhs, lhs
            
        sample_lhs = 1 * self._sample(lhs, subs)
        
        # short circuit if all zero
        if not np.any(sample_lhs):
            return sample_lhs
            
        sample_rhs = self._sample(rhs, subs)
        return sample_lhs * sample_rhs
    
    def _sample_product_grounded(self, args, subs):
        prod = 1
        
        # go through simple expressions first
        for arg in args: 
            if arg.is_constant_expression() or arg.is_pvariable_expression():
                sample = self._sample(arg, subs)
                prod *= sample
                if prod == 0:
                    return prod
        
        # go through complex expressions last
        for arg in args: 
            if not (arg.is_constant_expression() or arg.is_pvariable_expression()):
                sample = self._sample(arg, subs)
                prod *= sample
                if prod == 0:
                    return prod
                
        return prod
        
    # ===========================================================================
    # boolean
    # ===========================================================================
    
    def _sample_relational(self, expr, subs):
        _, op = expr.etype
        numpy_op = RDDLSimulator._check_op(
            op, self.RELATIONAL_OPS, 'Relational', expr)
        
        args = expr.args
        RDDLSimulator._check_arity(args, 2, op, expr)
        
        lhs, rhs = args
        sample_lhs = 1 * self._sample(lhs, subs)
        sample_rhs = 1 * self._sample(rhs, subs)
        return numpy_op(sample_lhs, sample_rhs)
    
    def _sample_logical(self, expr, subs):
        _, op = expr.etype
        if op == '&':
            op = '^'
        numpy_op = RDDLSimulator._check_op(op, self.LOGICAL_OPS, 'Logical', expr)
        
        args = expr.args
        n = len(args)
        
        if n == 1 and op == '~':
            arg, = args
            sample = self._sample(arg, subs)
            RDDLSimulator._check_type(sample, bool, op, expr, arg='')
            return np.logical_not(sample)
        
        # try to short-circuit ^ and | if possible
        elif n == 2:
            lhs, rhs = args
            if op == '^' or op == '|':
                return self._sample_and_or(lhs, rhs, op, expr, subs)
            else:
                sample_lhs = self._sample(lhs, subs)
                sample_rhs = self._sample(rhs, subs)
                RDDLSimulator._check_type(sample_lhs, bool, op, expr, arg=1)
                RDDLSimulator._check_type(sample_rhs, bool, op, expr, arg=2)
                return numpy_op(sample_lhs, sample_rhs)
        
        # for a grounded domain, we can short-circuit ^ and |
        elif n > 0 and (op == '^' or op == '|') \
        and not self.traced.cached_objects_in_scope(expr):
            return self._sample_and_or_grounded(args, op, expr, subs)
            
        raise RDDLInvalidNumberOfArgumentsError(
            f'Logical operator {op} does not have the required '
            f'number of arguments.\n' + print_stack_trace(expr))
    
    def _sample_and_or(self, lhs, rhs, op, expr, subs):
        
        # prioritize simple expressions
        if rhs.is_constant_expression() or rhs.is_pvariable_expression():
            lhs, rhs = rhs, lhs 
            
        sample_lhs = self._sample(lhs, subs)
        RDDLSimulator._check_type(sample_lhs, bool, op, expr, arg=1)
        
        # short-circuit if all True/False
        if (op == '^' and not np.any(sample_lhs)) \
        or (op == '|' and np.all(sample_lhs)):
            return sample_lhs
            
        sample_rhs = self._sample(rhs, subs)
        RDDLSimulator._check_type(sample_rhs, bool, op, expr, arg=2)
        
        if op == '^':
            return np.logical_and(sample_lhs, sample_rhs)
        else:
            return np.logical_or(sample_lhs, sample_rhs)
    
    def _sample_and_or_grounded(self, args, op, expr, subs): 
        use_and = op == '^'
        
        # go through simple expressions first
        for (i, arg) in enumerate(args):
            if arg.is_constant_expression() or arg.is_pvariable_expression():
                sample = self._sample(arg, subs)
                RDDLSimulator._check_type(sample, bool, op, expr, arg=i + 1)
                sample = bool(sample)
                if use_and and not sample:
                    return False
                elif not use_and and sample:
                    return True
        
        # go through complex expressions last
        for (i, arg) in enumerate(args):
            if not (arg.is_constant_expression() or arg.is_pvariable_expression()):
                sample = self._sample(arg, subs)
                RDDLSimulator._check_type(sample, bool, op, expr, arg=i + 1)
                sample = bool(sample)
                if use_and and not sample:
                    return False
                elif not use_and and sample:
                    return True
            
        return use_and
            
    # ===========================================================================
    # aggregation
    # ===========================================================================
    
    def _sample_aggregation(self, expr, subs):
        _, op = expr.etype
        numpy_op = RDDLSimulator._check_op(
            op, self.AGGREGATION_OPS, 'Aggregation', expr)
        
        # sample the argument and aggregate over the reduced axes
        * _, arg = expr.args
        sample = self._sample(arg, subs)                
        if op in self.AGGREGATION_BOOL:
            RDDLSimulator._check_type(sample, bool, op, expr, arg='')
        else:
            sample = 1 * sample
        _, axes = self.traced.cached_sim_info(expr)
        return numpy_op(sample, axis=axes)
     
    # ===========================================================================
    # function
    # ===========================================================================
    
    def _sample_func(self, expr, subs):
        _, name = expr.etype
        args = expr.args
        
        # unary function
        unary_op = self.UNARY.get(name, None)
        if unary_op is not None:
            RDDLSimulator._check_arity(args, 1, name, expr)
            arg, = args
            sample = 1 * self._sample(arg, subs)
            try:
                return unary_op(sample)
            except:
                raise ArithmeticError(
                    f'Cannot evaluate unary function {name} at {sample}.\n' + 
                    print_stack_trace(expr))
        
        # binary function
        binary_op = self.BINARY.get(name, None)
        if binary_op is not None:
            RDDLSimulator._check_arity(args, 2, name, expr)
            lhs, rhs = args
            sample_lhs = 1 * self._sample(lhs, subs)
            sample_rhs = 1 * self._sample(rhs, subs)
            if name in self.BINARY_REQUIRES_INT:
                RDDLSimulator._check_type(
                    sample_lhs, RDDLValueInitializer.INT, name, expr, arg=1)
                RDDLSimulator._check_type(
                    sample_rhs, RDDLValueInitializer.INT, name, expr, arg=2)
            try:
                return binary_op(sample_lhs, sample_rhs)
            except:
                raise ArithmeticError(
                    f'Cannot evaluate binary function {name} at '
                    f'{sample_lhs} and {sample_rhs}.\n' + print_stack_trace(expr))
        
        raise RDDLNotImplementedError(
            f'Function {name} is not supported.\n' + print_stack_trace(expr))
    
    # ===========================================================================
    # control flow
    # ===========================================================================
    
    def _sample_control(self, expr, subs):
        _, op = expr.etype
        RDDLSimulator._check_op(op, self.CONTROL_OPS, 'Control', expr)
        
        if op == 'if':
            return self._sample_if(expr, subs)
        else:
            return self._sample_switch(expr, subs)    
        
    def _sample_if(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 3, 'If then else', expr)
        
        pred, arg1, arg2 = args
        sample_pred = self._sample(pred, subs)
        RDDLSimulator._check_type(sample_pred, bool, 'If predicate', expr)
        
        # can short circuit if all elements of predicate tensor equal
        first_elem = bool(sample_pred.flat[0] 
                          if self.traced.cached_objects_in_scope(expr) 
                          else sample_pred)
        all_equal = np.all(sample_pred == first_elem)
        
        if all_equal:
            arg = arg1 if first_elem else arg2
            return self._sample(arg, subs)
        else:
            sample_then = self._sample(arg1, subs)
            sample_else = self._sample(arg2, subs)
            return np.where(sample_pred, sample_then, sample_else)
    
    def _sample_switch(self, expr, subs):
        pred, *_ = expr.args             
        sample_pred = self._sample(pred, subs)
        RDDLSimulator._check_type(
            sample_pred, RDDLValueInitializer.INT, 'Switch predicate', expr)
        
        # can short circuit if all elements of predicate tensor equal
        cases, default = self.traced.cached_sim_info(expr)  
        first_elem = bool(sample_pred.flat[0] 
                          if self.traced.cached_objects_in_scope(expr) 
                          else sample_pred)
        all_equal = np.all(sample_pred == first_elem)
        
        if all_equal:
            arg = cases[first_elem]
            if arg is None:
                arg = default
            return self._sample(arg, subs)        
        else: 
            sample_def = None if default is None else self._sample(default, subs)
            sample_cases = np.asarray([
                (sample_def if arg is None else self._sample(arg, subs))
                for arg in cases
            ])
            sample_pred = sample_pred[np.newaxis, ...]
            sample = np.take_along_axis(sample_cases, sample_pred, axis=0)   
            assert sample.shape[0] == 1
            return sample[0, ...]
        
    # ===========================================================================
    # random variables
    # ===========================================================================
    
    def _sample_random(self, expr, subs):
        _, name = expr.etype
        if name == 'KronDelta':
            return self._sample_kron_delta(expr, subs)        
        elif name == 'DiracDelta':
            return self._sample_dirac_delta(expr, subs)
        elif name == 'Uniform':
            return self._sample_uniform(expr, subs)
        elif name == 'Bernoulli':
            return self._sample_bernoulli(expr, subs)
        elif name == 'Normal':
            return self._sample_normal(expr, subs)
        elif name == 'Poisson':
            return self._sample_poisson(expr, subs)
        elif name == 'Exponential':
            return self._sample_exponential(expr, subs)
        elif name == 'Weibull':
            return self._sample_weibull(expr, subs)        
        elif name == 'Gamma':
            return self._sample_gamma(expr, subs)
        elif name == 'Binomial':
            return self._sample_binomial(expr, subs)
        elif name == 'NegativeBinomial':
            return self._sample_negative_binomial(expr, subs)
        elif name == 'Beta':
            return self._sample_beta(expr, subs)
        elif name == 'Geometric':
            return self._sample_geometric(expr, subs)
        elif name == 'Pareto':
            return self._sample_pareto(expr, subs)
        elif name == 'Student':
            return self._sample_student(expr, subs)
        elif name == 'Gumbel':
            return self._sample_gumbel(expr, subs)
        elif name == 'Laplace':
            return self._sample_laplace(expr, subs)
        elif name == 'Cauchy':
            return self._sample_cauchy(expr, subs)
        elif name == 'Gompertz':
            return self._sample_gompertz(expr, subs)
        elif name == 'ChiSquare':
            return self._sample_chisquare(expr, subs)
        elif name == 'Kumaraswamy':
            return self._sample_kumaraswamy(expr, subs)
        elif name == 'Discrete':
            return self._sample_discrete(expr, subs, unnorm=False)
        elif name == 'UnnormDiscrete':
            return self._sample_discrete(expr, subs, unnorm=True)
        elif name == 'Discrete(p)':
            return self._sample_discrete_pvar(expr, subs, unnorm=False)
        elif name == 'UnnormDiscrete(p)':
            return self._sample_discrete_pvar(expr, subs, unnorm=True)
        else:
            raise RDDLNotImplementedError(
                f'Distribution {name} is not supported.\n' + 
                print_stack_trace(expr))

    def _sample_kron_delta(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 1, 'KronDelta', expr)
        
        arg, = args
        sample = self._sample(arg, subs)
        RDDLSimulator._check_types(
            sample, (bool, RDDLValueInitializer.INT), 'Argument of KronDelta', expr)
        return sample
    
    def _sample_dirac_delta(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 1, 'DiracDelta', expr)
        
        arg, = args
        sample = self._sample(arg, subs)
        RDDLSimulator._check_type(
            sample, RDDLValueInitializer.REAL, 'Argument of DiracDelta', expr)        
        return sample
    
    def _sample_uniform(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Uniform', expr)

        lb, ub = args
        sample_lb = self._sample(lb, subs)
        sample_ub = self._sample(ub, subs)
        RDDLSimulator._check_bounds(sample_lb, sample_ub, 'Uniform', expr)
        return self.rng.uniform(low=sample_lb, high=sample_ub)      
    
    def _sample_bernoulli(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 1, 'Bernoulli', expr)
        
        pr, = args
        sample_pr = self._sample(pr, subs)
        RDDLSimulator._check_range(sample_pr, 0, 1, 'Bernoulli p', expr)
        size = sample_pr.shape if self.traced.cached_objects_in_scope(expr) else None
        return self.rng.uniform(size=size) <= sample_pr
    
    def _sample_normal(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Normal', expr)
        
        mean, var = args
        sample_mean = self._sample(mean, subs)
        sample_var = self._sample(var, subs)
        RDDLSimulator._check_positive(sample_var, False, 'Normal variance', expr)  
        sample_std = np.sqrt(sample_var)
        return self.rng.normal(loc=sample_mean, scale=sample_std)
    
    def _sample_poisson(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 1, 'Poisson', expr)
        
        rate, = args
        sample_rate = self._sample(rate, subs)
        RDDLSimulator._check_positive(sample_rate, False, 'Poisson rate', expr)        
        return self.rng.poisson(lam=sample_rate)
    
    def _sample_exponential(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 1, 'Exponential', expr)
        
        scale, = expr.args
        sample_scale = self._sample(scale, subs)
        RDDLSimulator._check_positive(sample_scale, True, 'Exponential rate', expr)
        return self.rng.exponential(scale=sample_scale)
    
    def _sample_weibull(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Weibull', expr)
        
        shape, scale = args
        sample_shape = self._sample(shape, subs)
        sample_scale = self._sample(scale, subs)
        RDDLSimulator._check_positive(sample_shape, True, 'Weibull shape', expr)
        RDDLSimulator._check_positive(sample_scale, True, 'Weibull scale', expr)
        return sample_scale * self.rng.weibull(a=sample_shape)
    
    def _sample_gamma(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Gamma', expr)
        
        shape, scale = args
        sample_shape = self._sample(shape, subs)
        sample_scale = self._sample(scale, subs)
        RDDLSimulator._check_positive(sample_shape, True, 'Gamma shape', expr)            
        RDDLSimulator._check_positive(sample_scale, True, 'Gamma scale', expr)        
        return self.rng.gamma(shape=sample_shape, scale=sample_scale)
    
    def _sample_binomial(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Binomial', expr)
        
        count, pr = args
        sample_count = self._sample(count, subs)
        sample_pr = self._sample(pr, subs)
        RDDLSimulator._check_type(sample_count, RDDLValueInitializer.INT, 'Binomial count', expr)
        RDDLSimulator._check_positive(sample_count, False, 'Binomial count', expr)
        RDDLSimulator._check_range(sample_pr, 0, 1, 'Binomial p', expr)
        return self.rng.binomial(n=sample_count, p=sample_pr)
    
    def _sample_negative_binomial(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'NegativeBinomial', expr)
        
        count, pr = args
        sample_count = self._sample(count, subs)
        sample_pr = self._sample(pr, subs)
        RDDLSimulator._check_positive(sample_count, True, 'NegativeBinomial r', expr)
        RDDLSimulator._check_range(sample_pr, 0, 1, 'NegativeBinomial p', expr)        
        return self.rng.negative_binomial(n=sample_count, p=sample_pr)
    
    def _sample_beta(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Beta', expr)
        
        shape, rate = args
        sample_shape = self._sample(shape, subs)
        sample_rate = self._sample(rate, subs)
        RDDLSimulator._check_positive(sample_shape, True, 'Beta shape', expr)
        RDDLSimulator._check_positive(sample_rate, True, 'Beta rate', expr)        
        return self.rng.beta(a=sample_shape, b=sample_rate)

    def _sample_geometric(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 1, 'Geometric', expr)
        
        pr, = args
        sample_pr = self._sample(pr, subs)
        RDDLSimulator._check_range(sample_pr, 0, 1, 'Geometric p', expr)        
        return self.rng.geometric(p=sample_pr)
    
    def _sample_pareto(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Pareto', expr)
        
        shape, scale = args
        sample_shape = self._sample(shape, subs)
        sample_scale = self._sample(scale, subs)
        RDDLSimulator._check_positive(sample_shape, True, 'Pareto shape', expr)        
        RDDLSimulator._check_positive(sample_scale, True, 'Pareto scale', expr)        
        return sample_scale * self.rng.pareto(a=sample_shape)
    
    def _sample_student(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 1, 'Student', expr)
        
        df, = args
        sample_df = self._sample(df, subs)
        RDDLSimulator._check_positive(sample_df, True, 'Student df', expr)            
        return self.rng.standard_t(df=sample_df)

    def _sample_gumbel(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Gumbel', expr)
        
        mean, scale = args
        sample_mean = self._sample(mean, subs)
        sample_scale = self._sample(scale, subs)
        RDDLSimulator._check_positive(sample_scale, True, 'Gumbel scale', expr)
        return self.rng.gumbel(loc=sample_mean, scale=sample_scale)
    
    def _sample_laplace(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Laplace', expr)
        
        mean, scale = args
        sample_mean = self._sample(mean, subs)
        sample_scale = self._sample(scale, subs)
        RDDLSimulator._check_positive(sample_scale, True, 'Laplace scale', expr)
        return self.rng.laplace(loc=sample_mean, scale=sample_scale)
    
    def _sample_cauchy(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Cauchy', expr)
        
        mean, scale = args
        sample_mean = self._sample(mean, subs)
        sample_scale = self._sample(scale, subs)
        RDDLSimulator._check_positive(sample_scale, True, 'Cauchy scale', expr)
        size = sample_mean.shape if self.traced.cached_objects_in_scope(expr) else None
        cauchy01 = self.rng.standard_cauchy(size=size)
        return sample_mean + sample_scale * cauchy01
    
    def _sample_gompertz(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Gompertz', expr)
        
        shape, scale = args
        sample_shape = self._sample(shape, subs)
        sample_scale = self._sample(scale, subs)
        RDDLSimulator._check_positive(sample_shape, True, 'Gompertz shape', expr)
        RDDLSimulator._check_positive(sample_scale, True, 'Gompertz scale', expr)
        size = sample_shape.shape if self.traced.cached_objects_in_scope(expr) else None
        U = self.rng.uniform(size=size)
        return np.log(1.0 - np.log1p(-U) / sample_shape) / sample_scale
    
    def _sample_chisquare(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 1, 'ChiSquare', expr)
        
        df, = args
        sample_df = self._sample(df, subs)
        RDDLSimulator._check_positive(sample_df, True, 'ChiSquare df', expr)
        return self.rng.chisquare(df=sample_df)
    
    def _sample_kumaraswamy(self, expr, subs):
        args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Kumaraswamy', expr)
        
        a, b = args
        sample_a = self._sample(a, subs)
        sample_b = self._sample(b, subs)
        RDDLSimulator._check_positive(sample_a, True, 'Kumaraswamy a', expr)
        RDDLSimulator._check_positive(sample_b, True, 'Kumaraswamy b', expr)
        size = sample_a.shape if self.traced.cached_objects_in_scope(expr) else None
        U = self.rng.uniform(size=size)
        return (1.0 - U ** (1.0 / sample_b)) ** (1.0 / sample_a)
    
    # ===========================================================================
    # random variables with enum support
    # ===========================================================================
    
    def _sample_discrete_helper(self, pdf, unnorm, expr):
        RDDLSimulator._check_positive(pdf, False, 'Discrete probabilities', expr)
        
        # calculate CDF       
        cdf = np.cumsum(pdf, axis=-1)
        if unnorm:
            cdf = cdf / cdf[..., -1:]
            
        # check valid CDF - still do this for unnorm to reject nan values
        if not np.allclose(cdf[..., -1], 1.0):
            raise RDDLValueOutOfRangeError(
                f'Discrete probabilities must sum to 1, got {cdf[..., -1]}.\n' + 
                print_stack_trace(expr))     
        
        # use inverse CDF sampling                  
        U = self.rng.random(size=cdf.shape[:-1] + (1,))
        return np.argmax(U < cdf, axis=-1)
        
    def _sample_discrete(self, expr, subs, unnorm):
        sorted_args = self.traced.cached_sim_info(expr)
        samples = [self._sample(arg, subs) for arg in sorted_args]
        pdf = np.stack(samples, axis=-1)
        return self._sample_discrete_helper(pdf, unnorm, expr)
    
    def _sample_discrete_pvar(self, expr, subs, unnorm):
        _, args = expr.args
        arg, = args
        pdf = self._sample(arg, subs)
        return self._sample_discrete_helper(pdf, unnorm, expr)
               
    # ===========================================================================
    # random vectors
    # ===========================================================================
    
    def _sample_random_vector(self, expr, subs):
        _, name = expr.etype
        if name == 'MultivariateNormal':
            return self._sample_multivariate_normal(expr, subs)
        elif name == 'MultivariateStudent':
            return self._sample_multivariate_student(expr, subs)
        elif name == 'Dirichlet':
            return self._sample_dirichlet(expr, subs)
        elif name == 'Multinomial':
            return self._sample_multinomial(expr, subs)
        else:
            raise RDDLNotImplementedError(
                f'Multivariate distribution {name} is not supported.\n' + 
                print_stack_trace(expr))
    
    def _sample_multivariate_normal(self, expr, subs):
        _, args = expr.args
        RDDLSimulator._check_arity(args, 2, 'MultivariateNormal', expr)
        
        mean, cov = args
        sample_mean = self._sample(mean, subs)
        sample_cov = self._sample(cov, subs)
        
        # reparameterization trick MN(m, LL') = LZ + m, where Z ~ Normal(0, 1)
        L = np.linalg.cholesky(sample_cov)
        Z = self.rng.standard_normal(
            size=sample_mean.shape + (1,),
            dtype=RDDLValueInitializer.REAL)
        sample = np.matmul(L, Z)[..., 0] + sample_mean

        # since the sampling is done in the last dimension we need to move it
        # to match the order of the CPF variables
        index, = self.traced.cached_sim_info(expr)
        sample = np.moveaxis(sample, source=-1, destination=index)
        return sample
    
    def _sample_multivariate_student(self, expr, subs):
        _, args = expr.args
        RDDLSimulator._check_arity(args, 3, 'MultivariateStudent', expr)
        
        mean, cov, df = args
        sample_mean = self._sample(mean, subs)
        sample_cov = self._sample(cov, subs)
        sample_df = self._sample(df, subs)
        RDDLSimulator._check_positive(sample_df, True, 'MultivariateStudent df', expr)
        
        # reparameterization trick MN(m, LL') = LZ + m, where Z ~ StudentT(0, 1)
        sample_df = sample_df[..., np.newaxis, np.newaxis]
        sample_df = np.broadcast_to(sample_df, shape=sample_mean.shape + (1,))
        L = np.linalg.cholesky(sample_cov)
        Z = self.rng.standard_t(df=sample_df)
        sample = np.matmul(L, Z)[..., 0] + sample_mean
        
        # since the sampling is done in the last dimension we need to move it
        # to match the order of the CPF variables
        index, = self.traced.cached_sim_info(expr)
        sample = np.moveaxis(sample, source=-1, destination=index)
        return sample
    
    def _sample_dirichlet(self, expr, subs):
        _, args = expr.args
        RDDLSimulator._check_arity(args, 1, 'Dirichlet', expr)
        
        alpha, = args
        sample_alpha = self._sample(alpha, subs)
        
        # sample Gamma(alpha_i, 1) and normalize across i
        RDDLSimulator._check_positive(sample_alpha, True, 'Dirichlet alpha', expr)
        Gamma = self.rng.gamma(shape=sample_alpha, scale=1.0)
        sample = Gamma / np.sum(Gamma, axis=-1, keepdims=True)
        
        # since the sampling is done in the last dimension we need to move it
        # to match the order of the CPF variables
        index, = self.traced.cached_sim_info(expr)
        sample = np.moveaxis(sample, source=-1, destination=index)
        return sample
    
    def _sample_multinomial(self, expr, subs):
        _, args = expr.args
        RDDLSimulator._check_arity(args, 2, 'Multinomial', expr)
        
        trials, prob = args
        sample_trials = self._sample(trials, subs) 
        sample_prob = self._sample(prob, subs)       
        RDDLSimulator._check_type(sample_trials, RDDLValueInitializer.INT, 'Multinomial trials', expr)
        RDDLSimulator._check_positive(sample_trials, False, 'Multinomial trials', expr)
        RDDLSimulator._check_positive(sample_prob, False, 'Discrete probabilities', expr)
        
        # check valid PMF
        cum_prob = np.sum(sample_prob, axis=-1)
        if not np.allclose(cum_prob, 1.0):
            raise RDDLValueOutOfRangeError(
                f'Multinomial probabilities must sum to 1, got {cum_prob}.\n' + 
                print_stack_trace(expr))    
            
        # sample from the multinomial
        sample = self.rng.multinomial(n=sample_trials, pvals=sample_prob)
        
        # since the sampling is done in the last dimension we need to move it
        # to match the order of the CPF variables
        index, = self.traced.cached_sim_info(expr)
        sample = np.moveaxis(sample, source=-1, destination=index)
        return sample
        
    # ===========================================================================
    # matrix algebra
    # ===========================================================================
    
    def _sample_matrix(self, expr, subs):
        _, op = expr.etype
        if op == 'det':
            return self._sample_matrix_det(expr, subs)
        elif op == 'inverse':
            return self._sample_matrix_inv(expr, subs, pseudo=False)
        elif op == 'pinverse':
            return self._sample_matrix_inv(expr, subs, pseudo=True)
        elif op == 'cholesky':
            return self._sample_matrix_cholesky(expr, subs)
        else:
            raise RDDLNotImplementedError(
                f'Matrix operator {op} is not supported.\n' + 
                print_stack_trace(expr))
    
    def _sample_matrix_det(self, expr, subs):
        * _, arg = expr.args
        sample_arg = self._sample(arg, subs)
        return np.linalg.det(sample_arg)
    
    def _sample_matrix_inv(self, expr, subs, pseudo):
        _, arg = expr.args
        sample_arg = self._sample(arg, subs)
        op = np.linalg.pinv if pseudo else np.linalg.inv
        sample = op(sample_arg)
        
        # matrix dimensions are last two axes, move them to the correct position
        indices = self.traced.cached_sim_info(expr)
        sample = np.moveaxis(sample, source=(-2, -1), destination=indices)
        return sample        
    
    def _sample_matrix_cholesky(self, expr, subs):
        _, arg = expr.args
        sample_arg = self._sample(arg, subs)
        op = np.linalg.cholesky
        sample = op(sample_arg)
        
        # matrix dimensions are last two axes, move them to the correct position
        indices = self.traced.cached_sim_info(expr)
        sample = np.moveaxis(sample, source=(-2, -1), destination=indices)
        return sample  
    
    
def lngamma(x):
    xmin = np.min(x)
    if not (xmin > 0):
        raise ValueError(f'Cannot evaluate log-gamma at {xmin}.')
    
    # small x: use lngamma(x) = lngamma(x + m) - ln(x + m - 1)... - ln(x)
    # large x: use asymptotic expansion OEIS:A046969
    if xmin < 7:
        return lngamma(x + 2) - np.log(x) - np.log(x + 1)        
    x_squared = x * x
    return (x - 0.5) * np.log(x) - x + 0.5 * np.log(2 * np.pi) + \
        1 / (12 * x) * (
            1 + 1 / (30 * x_squared) * (
                -1 + 1 / (7 * x_squared / 2) * (
                    1 + 1 / (4 * x_squared / 3) * (
                        -1 + 1 / (99 * x_squared / 140) * (
                            1 + 1 / (910 * x_squared / 3))))))


# A container class for compiling a simulator but from external info
class RDDLSimulatorPrecompiled(RDDLSimulator):
    
    def __init__(self, rddl: PlanningModel, 
                 init_values: Args, 
                 levels: Dict[int, Set[str]], 
                 trace_info: object,
                 rng: np.random.Generator=np.random.default_rng(),
                 keep_tensors: bool=False) -> None:
        self.init_values = init_values
        self.levels = levels
        self.traced = trace_info
        
        super(RDDLSimulatorPrecompiled, self).__init__(
            rddl=rddl, 
            allow_synchronous_state=True,
            rng=rng,
            logger=None,
            keep_tensors=keep_tensors)        
    
    def _compile(self):
        rddl = self.rddl
        
        # compute dependency graph for CPFs and sort them by evaluation order   
        self.cpfs = []  
        for cpfs in self.levels.values():
            for cpf in cpfs:
                _, expr = rddl.cpfs[cpf]
                prange = rddl.variable_ranges[cpf]
                dtype = RDDLValueInitializer.NUMPY_TYPES.get(
                    prange, RDDLValueInitializer.INT)
                self.cpfs.append((cpf, expr, dtype))
                
        # initialize all fluent and non-fluent values        
        self.subs = self.init_values.copy()
        self.state = None  
        self.noop_actions = {var: values
                             for (var, values) in self.init_values.items()
                             if rddl.variable_types[var] == 'action-fluent'}
        self._pomdp = bool(rddl.observ)
        
        # cached for performance
        self.invariant_names = [f'Invariant {i}' for i in range(len(rddl.invariants))]        
        self.precond_names = [f'Precondition {i}' for i in range(len(rddl.preconditions))]
        self.terminal_names = [f'Termination {i}' for i in range(len(rddl.terminals))]
        
        self.grounded_actionsranges = rddl.groundactionsranges()
        self.grounded_noop_actions = rddl.ground_values_from_dict(self.noop_actions)
        


'''This example runs hyper-parameter tuning on the Jax planner. The tuning
is performed using a batched parallelized Bayesian optimization.

The syntax is:

    python JaxTuningExample.py <domain> <instance> <method> [<trials>] [<iters>] [<workers>]
    
where:
    <domain> is the name of a domain located in the /Examples directory
    <instance> is the instance number
    <method> is either slp, drp or replan, as described in JaxExample.py
    <trials> is the number of trials to simulate when estimating the meta loss
    (defaults to 5)
    <iters> is the number of iterations of Bayesian optimization to perform
    (defaults to 20)
    <workers> is the number of parallel workers (i.e. batch size), which must
    not exceed the number of cores available on the machine (defaults to 4)
'''
import os
import sys

from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
from pyRDDLGym.Core.Jax.JaxParameterTuning import JaxParameterTuningDRP
from pyRDDLGym.Core.Jax.JaxParameterTuning import JaxParameterTuningSLP
from pyRDDLGym.Core.Jax.JaxParameterTuning import JaxParameterTuningSLPReplan
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import load_config
from pyRDDLGym.Examples.ExampleManager import ExampleManager


def main(domain, instance, method, trials=5, iters=20, workers=4):
    
    # set up the environment
    info = ExampleManager.GetEnvInfo(domain)    
    env = RDDLEnv.build(info, instance, enforce_action_constraints=True)
    
    # load the config file with planner settings
    abs_path = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(abs_path, 'JaxPlanConfigs', f'{domain}_{method}.cfg')
    planner_args, plan_args, train_args = load_config(config_path)
    
    # define algorithm to perform tuning
    if method == 'slp':
        tuning_class = JaxParameterTuningSLP        
    elif method == 'drp':
        tuning_class = JaxParameterTuningDRP    
    elif method == 'replan':
        tuning_class = JaxParameterTuningSLPReplan
    
    tuning = tuning_class(env=env,
                          train_epochs=train_args['epochs'],
                          timeout_training=train_args['train_seconds'],
                          eval_trials=trials,
                          planner_kwargs=planner_args,
                          plan_kwargs=plan_args,
                          num_workers=workers,
                          gp_iters=iters)
    
    best = tuning.tune(key=train_args['key'], filename=f'gp_{method}', 
                       save_plot=True)
    print(f'best parameters found: {best}')


if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) < 3:
        print('python JaxTuningExample.py <domain> <instance> <method> [<trials>] [<iters>] [<workers>]')
        exit(1)
    if args[2] not in ['drp', 'slp', 'replan']:
        print('<method> in [drp, slp, replan]')
        exit(1)
    kwargs = {'domain': args[0], 'instance': args[1], 'method': args[2]}
    if len(args) >= 4: kwargs['trials'] = int(args[3])
    if len(args) >= 5: kwargs['iters'] = int(args[4])
    if len(args) >= 6: kwargs['workers'] = int(args[5])
    main(**kwargs) 
    
